unit passhader;

interface
uses classes,sysutils;

type


TTokenType=(tkident,tkinteger,tkfloat,tkstring,tkeof,tkerror,
            tkprogram,tkunit,tkinterface,tkimplementation,tkuses,tkwith,tkdo,tkdiscard,tkprocedure,tkfunction,tkforward,tkbegin,tkend,tkasm,tkif,tkthen,tkelse,tkvar,tkvarying,tkparam,tktype,tkrecord,tkidiv,tkmod,tkshr,tkshl,tkand,tkor,tkxor,tknot,
            tkassignment,tkadd,tksub,tkmult,tkdiv,tkequal,tklequal,tkless,tkgequal,tkgreater,tknotequal,tkaddress,tkderef,tksemicolon,tkcomma,tkopenbracket,tkclosebracket,tkpoint,tkdblpoint,tkarrayopen,tkarrayclose,tkrange,
            tktypecast,tklast);
const

tokentext:array[tkident..tklast] of string =('','','','','','',
            'program','unit','interface','implementation','uses','with','do','discard','procedure','function','forward','begin','end','asm','if','then','else','var','varying','param','type','record','div','mod','shr','shl','and','or','xor','not',
            ':=','+','-','*','/','=','<=','<','>=','>','<>','@','^',';',',','(',')','.',':','[',']','..',
            '','');

type
TTokenSet=set of TTokenType;

ESyntaxError=class(Exception);

TScanner=class
private
 current_char,next_char:char;
 stream:TStream;
 function ReadChar:char;
 procedure ReadText;
 procedure FindTokenStart;
 procedure ReadSymbol;
 procedure ReadNum;
 procedure ReadHexNum;
 function ReadString:string;
public
 SourceLine:integer;
 SourcePos:integer;
 TokenString:string;
 TokenType:TTokenType;
 ParsedText:string;
 procedure CheckToken(token:TTokenSet);
 procedure AcceptToken(token:TTokentype);
 function ReadToken:string;
 constructor create(st:TStream);
 destructor destroy;override;
end;

type

TPascalShader=class;

TSymbolScope=(scvar,scparam,scattrib,scoutput);

TSymbolInfo=class
 name:string;
 scope:TSymbolScope;
 arraysize:integer;
 offset:integer;
 temp,_default:boolean;
 varying:boolean;
 used:boolean;
end;

TTokenInfo=record
 token:TTokenType;
 SourcePos,SourceLine:integer;
 tokenstring:string;
end;

TExpression=class;
TStatement=class;

TFunctionInfo=class
 name:string;
 paramlist:array of string;
 codeblock:TStatement;
 called:boolean;
end;


TExpression=class
private
 info:TTokenInfo;
 ident:string;
 subnodes:array[0..1] of TExpression;
 paramlist:array of TExpression;
 symbolinfo:TSymbolInfo;
 satfunc:boolean;
 unary:boolean;
 lvalue:boolean;
 checked:boolean;
 shader:TPascalShader;
 procedure Push;
 function GetOperatorPrecedence:integer;
 function CompareOperator(exp:TExpression):boolean;
 procedure AddBinaryOperator;
 procedure AddUnaryOperator;
 procedure AddOperator;
 procedure CheckIdent;
 procedure CheckParamCount(c:integer);
 function IsConstVector:boolean;
 function OptimizeConstExpr:TExpression;
 function OptimizeMAD:TExpression;
 function OptimizeMul:TExpression;
 function OptimizeSub:TExpression;
 function OptimizeDiv:TExpression;
 function Optimize:TExpression;
 function OptimizeFunc:TExpression;
 function Calc(p1,p2:string):string;
 function Sat:string;
 procedure WriteCmd(const cmd,p1,p2,p3:string;const p4:string='');
 function AllocateTempVar:string;
 procedure FreeTempVar(const s:string);
 function WriteBinaryOperator(const target:string=''):string;
 function WriteUnaryOperator(const target:string=''):string;
 function WriteOperator(const target:string=''):string;
 function WriteExpressionIdent(const target:string=''):string;
 function WriteExpressionConst(const target:string=''):string;
 function WriteExpressionFunc(const target:string=''):string;
 function WriteExpressionFunc_OpCode1(const opcode:string;const target:string=''):string;
 function WriteExpressionFunc_OpCode1s(const opcode:string;const target:string=''):string;
 function WriteExpressionFunc_OpCode2(const opcode:string;const target:string=''):string;
 function WriteExpressionFunc_OpCode2s(const opcode:string;const target:string=''):string;
 function WriteExpressionFunc_rcp(const target:string=''):string;
 function WriteExpressionFunc_Normalize(const target:string=''):string;
 function WriteExpressionFunc_Length(const target:string=''):string;
 function WriteExpressionFunc_SqrLength(const target:string=''):string;
 function WriteExpressionFunc_InvLength(const target:string=''):string;
 function WriteExpressionFunc_Dist(const target:string=''):string;
 function WriteExpressionFunc_InvDist(const target:string=''):string;
 function WriteExpressionFunc_SqrDist(const target:string=''):string;
 function WriteExpressionFunc_Sqr(const target:string=''):string;
 function WriteExpressionFunc_Sqrt(const target:string=''):string;
 function WriteExpressionFunc_Mad(const target:string=''):string;
 function WriteExpressionFunc_MatrixMult3(const target:string=''):string;
 function WriteExpressionFunc_MatrixMult4(const target:string=''):string;
 function WriteExpressionFunc_Vector(const target:string=''):string;
 function WriteExpressionFunc_sat(const target:string=''):string;
 function WriteExpressionFunc_clamp(const target:string=''):string;
 function WriteExpressionFunc_reflect(const target:string=''):string;
 function WriteExpressionFunc_lightatten(const target:string=''):string;
 function WriteExpressionFunc_Decl(func:TFunctionInfo;const target:string=''):string;
 function WriteExpression(target:string=''):string;
 constructor create(ps:TPascalShader);
end;

TStatementType=(stblock,stassignment,stwith,stdiscard,stif);

TStatement=class
private
 function AllocateTempVar:string;
 procedure FreeTempVar(const s:string);
 procedure WriteCmd(const cmd,p1,p2,p3:string;const p4:string='');
 procedure WriteAssignment;
 procedure WriteCodeBlock;
 procedure WriteWithStatement;
 procedure WriteIfStatement;
 procedure WriteDiscard;
public
 info:TTokenInfo;
 _type:TStatementType;
 exp1,exp2:TExpression;
 statements:array of TStatement;
 withdecl:string;
 shader:TPascalShader;
 procedure Write;
 constructor create(ps:TPascalShader);
end;

TStack=class
private
 fcount:integer;
 flist:TList;
public
 property Count:Integer read fcount;
 procedure Push(p:pointer);
 function Pop:Pointer;
 function Peek:Pointer;
 constructor create;
 destructor destroy;override;
end;

TARBCommand=class
 cmd:string;
 p1,p2,p3,p4:string;
 shader:TPascalShader;
function IsVarRead(const v:string):boolean;
function IsVarWritten(const v:string):boolean;
function IsVarOverWritten(const v:string):boolean;
function IsVarCompWritten(const v,comp:string):boolean;
constructor create(ps:TPascalShader);
end;

TPascalShaderClass=class of TPascalShader;

TPascalShader=class
private
 symbollist,nodelist,funclist:TList;
 expstack:TStack;
 exptree:TExpression;
 sc:TScanner;
 codeblock:TStatement;
 tempvarlist:array of string;
 paramsize:integer;
 matrixlist:TStringList;
 withlist:TStringList;
 withlistart:integer;
 fsource:string;
 cmdlist:array of TARBCommand;
 reservedwordlist:TStringList;
 SavedDecimalSeparator:char;
 currentcondition,currentconditioninv:string;
 currentcondition_cmp:string;
 currentfunc:string;
 function ReadTokenInfo:TTokenInfo;
 function ParseExpression_r:TExpression;
 function ParseExpression:TExpression;
 function ParseExpressionStatement:TStatement;
 function ParseStatement:TStatement;
 procedure ParseProgram;
 procedure ParseVarBlock(fd:TFunctionInfo);
 procedure ParseVarDecl(fd:TFunctionInfo);
 procedure ParseParamBlock;
 procedure ParseParamDecl;
 procedure ParseVaryingBlock;
 procedure ParseVaryingDecl;
 procedure ParseFuncDecl;
 function ParseWith:TStatement;
 function ParseDiscard:TStatement;
 function Parseif:TStatement;
 function ParseCodeBlock:TStatement;
 function FindValidMatrix(const s:string):string;
 function IsValidWithDecl(const s:string):boolean;
 function FindValidWithDecl(const s:string):string;
 procedure WriteLine(const s:string);
 procedure WriteCmd(const cmd,p1,p2,p3:string;const p4:string='');
 function FindSymbol(const sname:string):TSymbolInfo;
 function FindFunction(const fname:string):TFunctionInfo;
 procedure WriteProgram;
 procedure WriteVarBlock;
 procedure WriteVaryingBlock;
 procedure WriteParamBlock;
 procedure WriteCommandList;
 function RemoveDblAssignments:boolean;
 function RemoveDblCmd:boolean;
 function RemoveAssignments:boolean;
 function OptimizeCMP:boolean;
 function OptimizeSLT:boolean;
 function OptimizeSUB:boolean;
 procedure FindUsedSymbols;
 procedure OptimizeCommandList;
 function CanVarDiscarded(const vn:string;cmdid:integer):boolean;
 function FindNextRead(const vn:string;cmdid:integer):integer;
 procedure AddDefaultSymbol(const s:string;sc:TSymbolScope);
 function AllocateTempVar:string;
 procedure FreeTempVar(const s:string);
 procedure InitMatrixList;
 procedure InitReservedWords;
protected
 AllowDiscard:Boolean;
 SAT_Cmd:boolean;
 CMP_Cmd:boolean;
 VaryingType:TSymbolScope;
 Header:String;
 procedure InitDefaultVars;virtual;
 function WriteExpressionFunction(exp:TExpression;const target:string=''):string;virtual;
public
 property Source:string read fsource;
 function GetSymbolOffset(const v:string):integer;
 constructor create(st:TStream);virtual;
 destructor destroy;override;
end;

TVertexProgram=class(TPascalShader)
protected
 function WriteExpressionFunc_Cos(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_Sin(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_lerp(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunction(exp:TExpression;const target:string=''):string;override;
 procedure InitDefaultVars;override;
public
 constructor create(st:TStream);override;
end;

TFragmentProgram=class(TPascalShader)
protected
 function WriteExpressionFunc_texture(exp:TExpression;const texture:string;const target:string=''):string;
 function WriteExpressionFunc_textureproj(exp:TExpression;const texture:string;const target:string=''):string;
 function WriteExpressionFunc_normalmap(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_lerp(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunction(exp:TExpression;const target:string=''):string;override;
 procedure InitDefaultVars;override;
public
 constructor create(st:TStream);override;
end;

implementation

const
binaryop=[tkadd,tksub,tkmult,tkdiv,tkequal,tklequal,tkless,tkgequal,tkgreater,tknotequal,tkpoint,tkand,tkor,tkxor];
unaryop=[tkadd,tksub,tknot];
expressionstart=[tkident,tkinteger,tkfloat]+unaryop;

function TokenInfoStr(const ti:TTokenInfo):string;
begin
 result:='Line '+inttostr(ti.SourceLine)+':';
end;

function IsValidRSwizzle(const s:string):boolean;
var
 a,b,i:integer;
begin
a:=0;
b:=0;
for i:=1 to length(s) do
case Upcase(s[i]) of
 'X','Y','Z','W':inc(a);
 'R','G','B','A':inc(b);
 else
  begin
   result:=false;
   exit;
  end; 
end;
result:=((a+b) in [1,4]) and (length(s) in [1,4]);
end;

function IsValidLSwizzle(const s:string):boolean;
var
 a,b,c,d,i:integer;
begin
a:=0;
b:=0;
c:=0;
d:=0;
for i:=1 to length(s) do
case UpCase(s[i]) of
 'X','R':inc(a);
 'Y','G':inc(b);
 'Z','B':inc(c);
 'W','A':inc(d);
 else
  begin
   result:=false;
   exit;
  end;
end;
result:=((a<=1) and (b<=1) and (c<=1) and (d<=1)) and (a+b+c+d>0);
end;

function GetSwizzleMask(const s:string):string;
var
 p:integer;
begin
p:=pos('.',s);
if (p=0) or (pos('}',ExtractFileExt(s))<>0) then
begin
result:='';
exit;
end;
result:=ExtractFileExt(s);
result:=copy(result,2,length(result)-1);
if not (IsValidRSwizzle(result) or IsValidLSwizzle(result)) then
 result:='';
end;


function CorrectSwizzle(const s:string):string;
var
 sw,v:string;
begin
sw:=GetSwizzleMask(s);
if sw='' then
begin
 result:=s;
end
else
begin
sw:=LowerCase(sw);
sw:=StringReplace(sw,'r','x',[rfreplaceall]);
sw:=StringReplace(sw,'g','y',[rfreplaceall]);
sw:=StringReplace(sw,'b','z',[rfreplaceall]);
sw:=StringReplace(sw,'a','w',[rfreplaceall]);
v:=copy(s,1,length(s)-length(sw)-1);
result:=v+'.'+sw;
end;
end;


function ChangeScalarSwizzle(const s:string):string;
var
 v,sw:string;
begin
sw:=GetSwizzleMask(s);
if sw='' then
 v:=s
else
 v:=copy(s,1,length(s)-length(sw)-1);
if sw='' then
 result:=v+'.x'
else
 result:=v+'.'+sw[1];
end;

function CombineSwizzle(const s,sw:string):string;
var
 v,sw2,sw3:string;
 i:integer;
begin
sw2:=lowercase(GetSwizzleMask(s));
if sw2='' then
 v:=s
else
v:=copy(s,1,length(s)-length(sw2)-1);
if sw2='' then sw2:='xyzw';
 sw3:='';
for i:=1 to length(sw) do
 if pos(lowercase(sw[i]),sw2)<>0 then
  sw3:=sw3+lowercase(sw[i]);
if sw3='' then
 result:=''
else
 result:=v+'.'+sw3;
end;

function ExtractVarname(const s:string):string;
var
 sw:string;
begin
sw:=lowercase(GetSwizzleMask(s));
if sw='' then
 result:=s
else
 result:=copy(s,1,length(s)-length(sw)-1);
if copy(result,1,1)='-' then
 result:=copy(result,2,length(result)-1);
end;

function ChangeVarname(const s,n:string):string;
var
 sw,p:string;
begin
sw:=lowercase(GetSwizzleMask(s));
if sw<>'' then sw:='.'+sw;
if copy(s,1,1)='-' then
begin
 if copy(n,1,1)='-' then p:='' else p:='-';
end
else
 p:='';
result:=p+n+sw;
end;

function NegateStr(const s:string):string;
begin
result:=trim(s);
if copy(s,1,1)='-' then
 result:=copy(result,2,length(result)-1)
else
 result:='-'+result;
end;


// TStack

procedure TStack.Push(p:pointer);
begin
inc(fcount);
flist.Add(p);
end;

function TStack.Pop:Pointer;
begin
if count=0 then
 result:=nil
else
begin
 dec(fcount);
 result:=flist.Items[count];
 flist.Delete(count);
end;
end;

function TStack.Peek:Pointer;
begin
if count=0 then
 result:=nil
else
 result:=flist.Items[count-1];
end;

constructor TStack.create;
begin
inherited;
flist:=TList.create;
end;

destructor TStack.destroy;
begin
flist.free;
inherited;
end;


// TScanner

function TScanner.ReadChar:char;
begin
current_char:=next_char;
if stream.Position>=stream.Size then
 next_char:=#0
else
begin
 stream.Read(next_char,1);
 ParsedText:=ParsedText+next_char;
 inc(SourcePos);
end;
if current_char=#10 then inc(sourceline);
result:=current_char;
end;

procedure TScanner.FindTokenStart;
begin
while current_char in [' ',#13,#10] do ReadChar;
if current_char='{' then
begin
 while not (ReadChar in ['}',#0]) do;
end;
end;

procedure TScanner.ReadText;
var
 t:TTokenType;
begin
TokenType:=tkident;
TokenString:=current_char;
while ReadChar in ['A'..'Z','a'..'z','_','0'..'9'] do TokenString:=TokenString+current_char;
for t:=tkident to tklast do
 if sametext(Tokentext[t],TokenString) then
 begin
  TokenType:=t;
  break;
 end;
end;

procedure TScanner.ReadSymbol;
begin
TokenString:=current_char;
TokenType:=tkerror;
case current_char of
 '+':TokenType:=tkadd;
 '-':TokenType:=tksub;
 '*':TokenType:=tkmult;
 '/':
  begin
  if next_char='/' then
  begin
   ReadChar;
   TokenString:='//';
  end;

  TokenType:=tkdiv;
  end;
 ';':TokenType:=tksemicolon;
 ',':TokenType:=tkcomma;
 '(':TokenType:=tkopenbracket;
 ')':TokenType:=tkclosebracket;
 '.':
  begin
   if next_char='.' then
   begin
    TokenType:=tkrange;
    ReadChar;
    TokenString:='..';
   end
   else
   begin
    TokenType:=tkpoint;
   end;
  end;
 '=':TokenType:=tkequal;
 '@':TokenType:=tkaddress;
 '^':TokenType:=tkderef;
 '[':TokenType:=tkarrayopen;
 ']':TokenType:=tkarrayclose;
 '<':
  begin
   if next_char='=' then
   begin
    ReadChar;
    TokenString:='<=';
    TokenType:=tklequal;
   end
   else
   if next_char='>' then
   begin
    ReadChar;
    TokenString:='<>';
    TokenType:=tknotequal;
   end
   else
   begin
    TokenType:=tkless;
   end;
  end;
 '>':
  begin
   if next_char='=' then
   begin
    ReadChar;
    TokenString:='>=';
    TokenType:=tkgequal;
   end
   else
   begin
    TokenType:=tkgreater;
   end;
  end;
 ':':
  begin
   if next_char='=' then
   begin
    ReadChar;
    TokenString:=':=';
    TokenType:=tkassignment;
   end
   else
   begin
    TokenType:=tkdblpoint;
   end;
  end;
end;
ReadChar;
end;


procedure TScanner.ReadNum;
begin
Tokenstring:=current_char;
TokenType:=tkinteger;
while (ReadChar in ['0'..'9','.']) do
begin
 if (current_char='.') and (next_char='.') then break;
 if (current_char='.') then
 begin
  if TokenType=tkinteger then tokentype:=tkfloat else tokentype:=tkerror;
 end;
 Tokenstring:=Tokenstring+current_char;
end;

end;

procedure TScanner.ReadHexNum;
begin
Tokenstring:=current_char;
TokenType:=tkinteger;
while (ReadChar in ['0'..'9','A'..'F','a'..'f']) do
begin
 Tokenstring:=Tokenstring+current_char;
end;

end;


function TScanner.ReadString:string;
begin
while not (ReadChar in ['''',#13,#10,#0]) do TokenString:=TokenString+current_char;
if current_char='''' then TokenType:=tkstring else TokenType:=tkerror;
ReadChar;
end;

function TScanner.ReadToken:string;
begin
FindTokenStart;
TokenString:='';
TokenType:=tkerror;
case current_char of
 'A'..'Z','a'..'z','_':ReadText;
 '0'..'9':ReadNum;
 '$':ReadHexNum;
 '''':ReadString;
 #0:
 begin
  result:=#0;
  TokenType:=tkeof;
 end
else
 ReadSymbol;
if TokenString='//' then
begin
while not (ReadChar in [#13,#10,#0]) do;
TokenString:=ReadToken;
end;
end;
result:=TokenString;
end;

procedure TScanner.CheckToken(token:TTokenSet);
begin
if not (TokenType in token) then raise ESyntaxError.Create('Invalid Token');
end;

procedure TScanner.AcceptToken(token:TTokentype);
begin
CheckToken([token]);
ReadToken;
end;


constructor TScanner.create(st:TStream);
begin
inherited create;
SourceLine:=1;
SourcePos:=2;
stream:=st;
stream.ReadBuffer(current_char,1);
stream.ReadBuffer(next_char,1);
ParsedText:=current_char+next_char;
ReadToken;
end;

destructor TScanner.destroy;
begin
inherited;
end;

// TExpression

procedure TExpression.Push;
begin
shader.expstack.Push(self);
shader.exptree:=self;
end;

function TExpression.GetOperatorPrecedence:integer;
begin
result:=-1;
case info.token of
 tkpoint:result:=5;
 tknot,tkaddress:result:=4;
 tkmult,tkdiv,tkidiv,tkmod,tkand,tkshl,tkshr:result:=3;
 tkadd,tksub,tkor,tkxor:if unary then result:=4 else result:=2;
 tkequal,tknotequal,tklequal,tkless,tkgreater,tkgequal:result:=1;
end;
end;

function TExpression.CompareOperator(exp:TExpression):boolean;
begin
result:=GetOperatorPrecedence>=exp.GetOperatorPrecedence;
end;

procedure TExpression.AddBinaryOperator;
begin
if shader.expstack.count<2 then
begin
 raise ESyntaxError.Create(TokenInfoStr(info)+'Expression expected but "'+shader.sc.TokenString+'" found');
end;
subnodes[1]:=TExpression(shader.expstack.pop);
subnodes[0]:=TExpression(shader.expstack.pop);
Push;
end;

procedure TExpression.AddUnaryOperator;
begin
if shader.expstack.count<1 then
begin
 raise ESyntaxError.Create(TokenInfoStr(info)+'Expression expected but "'+shader.sc.TokenString+'" found');
end;
subnodes[0]:=TExpression(shader.expstack.pop);
Push;
end;

procedure TExpression.AddOperator;
begin
if unary then AddUnaryOperator else AddBinaryOperator;
end;

procedure TExpression.CheckIdent;
var
 v,s:string;
 b:boolean;
begin
if info.token=tkident then
begin
 if checked then exit;
 checked:=true;
 ident:=CorrectSwizzle(ident);
 s:=GetSwizzleMask(ident);
 if s='' then
  v:=ident
 else
  v:=copy(ident,1,length(ident)-length(s)-1);
 symbolinfo:=shader.FindSymbol(v);
 if symbolinfo=nil then
  raise ESyntaxError.Create(TokenInfoStr(info)+'Undeclared Identifier: "'+ident+'"');
 if s='' then
  ident:=symbolinfo.name
 else
 ident:=symbolinfo.name+'.'+s;
 if not lvalue and (symbolinfo.scope=scoutput) then
  raise ESyntaxError.Create(TokenInfoStr(info)+'Variable ist writeonly: "'+ident+'"');
 if s<>'' then
 begin
  if lvalue then b:=IsValidLSwizzle(s) else b:=IsValidRSwizzle(s);
  if not b then
   raise ESyntaxError.Create(TokenInfoStr(info)+'Invalid swizzle: "'+s+'"');
 end;
end
else
 raise ESyntaxError.Create(TokenInfoStr(info)+'Identifier expected ');
end;

procedure TExpression.CheckParamCount(c:integer);
begin
if length(paramlist)<>c then
 raise ESyntaxError.Create(tokeninfostr(info)+'Invalid parameter count');
end;

function TExpression.IsConstVector:boolean;
begin
result:=sametext(ident,'vector') and (length(paramlist)=4) and
        (paramlist[0].info.token in [tkinteger,tkfloat]) and
        (paramlist[1].info.token in [tkinteger,tkfloat]) and
        (paramlist[2].info.token in [tkinteger,tkfloat]) and
        (paramlist[3].info.token in [tkinteger,tkfloat]);
end;

function TExpression.Calc(p1,p2:string):string;
begin
result:='';
case info.token of
 tkadd:result:=floattostr(strtofloat(p1)+strtofloat(p2));
 tksub:result:=floattostr(strtofloat(p1)-strtofloat(p2));
 tkmult:result:=floattostr(strtofloat(p1)*strtofloat(p2));
 tkdiv:result:=floattostr(strtofloat(p1)/strtofloat(p2));
end;    
end;

function TExpression.OptimizeConstExpr:TExpression;
var
 i:integer;
 op1scalar,op2scalar,op1vec,op2vec:boolean;
begin
result:=self;

if not unary and (info.token in [tkadd,tksub,tkmult,tkdiv]) then
begin

op1scalar:=assigned(subnodes[0]) and (subnodes[0].info.token in [tkinteger,tkfloat]);
op2scalar:=assigned(subnodes[1]) and (subnodes[1].info.token in [tkinteger,tkfloat]);
op1vec:=assigned(subnodes[0]) and subnodes[0].IsConstVector;
op2vec:=assigned(subnodes[1]) and subnodes[1].IsConstVector;

if op1scalar then
begin
 if op2scalar then
 begin
  info.tokenstring:=Calc(subnodes[0].info.tokenstring,subnodes[1].info.tokenstring);
  info.token:=tkfloat;
 end;
 if op2vec then
 begin
  setlength(paramlist,4);
  for i:=0 to 3 do
  begin
   paramlist[i]:=subnodes[1].paramlist[i];
   paramlist[i].info.tokenstring:=calc(subnodes[0].info.tokenstring,subnodes[1].paramlist[i].info.tokenstring);
  end;
  info.token:=tkident;
  ident:='vector';
  info.tokenstring:='vector';
 end;
end;
if op1vec then
begin
 if op2scalar then
 begin
  setlength(paramlist,4);
  for i:=0 to 3 do
  begin
   paramlist[i]:=subnodes[0].paramlist[i];
   paramlist[i].info.tokenstring:=calc(subnodes[0].paramlist[i].info.tokenstring,subnodes[1].info.tokenstring);
  end;
  info.token:=tkident;
  ident:='vector';
  info.tokenstring:='vector';
 end;
 if op2vec then
 begin
  setlength(paramlist,4);
  for i:=0 to 3 do
  begin
   paramlist[i]:=subnodes[0].paramlist[i];
   paramlist[i].info.tokenstring:=calc(subnodes[0].paramlist[i].info.tokenstring,subnodes[1].paramlist[i].info.tokenstring);
  end;
  info.token:=tkident;
  ident:='vector';
  info.tokenstring:='vector';
 end;
end;
end;
end;

function TExpression.OptimizeMAD:TExpression;
begin
result:=self;

if not unary and (info.token=tkadd) then
begin
if subnodes[0].info.token=tkmult then
begin
 info.token:=tkident;
 ident:='mad';
 setlength(paramlist,3);
 paramlist[0]:=subnodes[0].subnodes[0];
 paramlist[1]:=subnodes[0].subnodes[1];
 paramlist[2]:=subnodes[1];
end
else
if subnodes[1].info.token=tkmult then
begin
 info.token:=tkident;
 ident:='mad';
 setlength(paramlist,3);
 paramlist[0]:=subnodes[1].subnodes[0];
 paramlist[1]:=subnodes[1].subnodes[1];
 paramlist[2]:=subnodes[0];
end;
end;

end;

function TExpression.OptimizeMul:TExpression;
begin
result:=self;
if not unary and (info.token=tkmult) then
begin
if (subnodes[0].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(subnodes[0].info.tokenstring)-1)<0.00000001) then
begin
  result:=subnodes[1];
end
else
if (subnodes[1].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(subnodes[1].info.tokenstring)-1)<0.00000001) then
begin
  result:=subnodes[0];
end
else
if (subnodes[0].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(subnodes[0].info.tokenstring))<0.00000001) then
begin
 info.token:=tkinteger;
 info.tokenstring:='0';
end
else
if (subnodes[1].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(subnodes[1].info.tokenstring))<0.00000001) then
begin
 info.token:=tkinteger;
 info.tokenstring:='0';
end;
end;
end;

function TExpression.OptimizeSub:TExpression;
begin
result:=self;
if unary and (info.token=tksub) then
begin
 if (subnodes[0].info.token in [tkinteger,tkfloat])then
 begin
  info.token:=tkfloat;
  info.tokenstring:=floattostr(-strtofloat(subnodes[0].info.tokenstring));
 end;
end;
end;

function TExpression.OptimizeDiv:TExpression;
var
 i:integer;
begin
result:=self;
if (info.token=tkdiv) then
begin
if (subnodes[0].info.token in [tkinteger,tkfloat]) and (subnodes[1].info.token in [tkinteger,tkfloat]) then
begin
 info.token:=tkfloat;
 info.tokenstring:=floattostr(strtofloat(subnodes[0].info.tokenstring)/strtofloat(subnodes[1].info.tokenstring));
end
else
if (subnodes[1].info.token in [tkinteger,tkfloat]) then
begin
 info.token:=tkmult;
 subnodes[1].info.tokenstring:=floattostr(1/strtofloat(subnodes[1].info.tokenstring));
 result:=Optimize;
end
else
if subnodes[1].IsConstVector then
begin
 info.token:=tkmult;
 for i:=0 to 3 do
  subnodes[1].paramlist[i].info.tokenstring:=floattostr(1/strtofloat(subnodes[1].paramlist[i].info.tokenstring));
 result:=Optimize;
end
else
if (subnodes[0].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(subnodes[0].info.tokenstring))<0.00000001) then
begin
 info.token:=tkinteger;
 info.tokenstring:='0';
end
else
if (subnodes[1].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(subnodes[1].info.tokenstring))<0.00000001) then
begin
 info.token:=tkinteger;
 info.tokenstring:='0';
end
else
if sametext(subnodes[1].info.tokenstring,'sqrt')   then
begin
 info.token:=tkmult;
 subnodes[1].info.tokenstring:='rsq';
 subnodes[1].ident:='rsq';
 result:=Optimize;
end
else
if sametext(subnodes[1].info.tokenstring,'length') then
begin
 info.token:=tkmult;
 subnodes[1].info.tokenstring:='invlength';
 subnodes[1].info.token:=tkident;
 subnodes[1].ident:='invlength';
 result:=Optimize;
end
else
if sametext(subnodes[1].info.tokenstring,'dist') then
begin
 info.token:=tkmult;
 subnodes[1].info.tokenstring:='invdist';
 subnodes[1].info.token:=tkident;
 subnodes[1].ident:='invdist';
 result:=Optimize;
end
else
if (subnodes[0].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(subnodes[0].info.tokenstring)-1)<0.00000001) then
begin
 info.token:=tkident;
 info.tokenstring:='rcp';
 ident:='rcp';
 setlength(paramlist,1);
 paramlist[0]:=subnodes[1];
 result:=Optimize;
end;
end;
end;


function TExpression.OptimizeFunc:TExpression;
begin
result:=self;
if info.token=tkident then
begin
if SameText(ident,'pi') then
begin
 info.token:=tkfloat;
 info.tokenstring:='3.1415926535';
end;
if SameText(ident,'sqr') and (length(paramlist)=1) then
 if (paramlist[0].info.token=tkident) and SameText(paramlist[0].ident,'length') then
 begin
  paramlist[0].ident:='sqrlength';
  result:=result.Optimize;
 end;
if SameText(ident,'sqr') and (length(paramlist)=1) then
 if (paramlist[0].info.token=tkident) and SameText(paramlist[0].ident,'dist') then
 begin
  paramlist[0].ident:='sqrdist';
  result:=result.Optimize;
 end;
if SameText(ident,'sqrt') and (length(paramlist)=1) then
 if (paramlist[0].info.token in [tkinteger,tkfloat]) then
 begin
  info.token:=tkfloat;
  info.tokenstring:=floattostr(sqrt(strtofloat(paramlist[0].info.tokenstring)));
 end;
if SameText(ident,'sqr') and (length(paramlist)=1) then
 if (paramlist[0].info.token in [tkinteger,tkfloat]) then
 begin
  info.token:=tkfloat;
  info.tokenstring:=floattostr(sqr(strtofloat(paramlist[0].info.tokenstring)));
 end;
if SameText(ident,'rsqrt') and (length(paramlist)=1) then
 if (paramlist[0].info.token in [tkinteger,tkfloat]) then
  begin
   info.token:=tkfloat;
   info.tokenstring:=floattostr(1/sqrt(strtofloat(paramlist[0].info.tokenstring)));
  end;
end;

if shader.SAT_Cmd and (SameText(ident,'sat') and (length(paramlist)=1)) then
 begin
  paramlist[0].satfunc:=true;
  result:=paramlist[0].Optimize;
 end;


end;

function TExpression.Optimize:TExpression;
var
 i:integer;
begin
if assigned(subnodes[0]) then subnodes[0]:=subnodes[0].Optimize;
if assigned(subnodes[1]) then subnodes[1]:=subnodes[1].Optimize;
for i:=0 to length(paramlist)-1 do
 paramlist[i]:=paramlist[i].Optimize;
self:=OptimizeConstExpr;
self:=OptimizeMAD;
self:=OptimizeMUL;
self:=OptimizeSub;
self:=OptimizeDiv;
self:=OptimizeFunc;
result:=self;
end;

function TExpression.Sat:string;
begin
if satfunc then result:='_SAT' else result:='';
end;

procedure TExpression.WriteCmd(const cmd,p1,p2,p3:string;const p4:string='');
begin
shader.WriteCmd(cmd,p1,p2,p3,p4);
end;

function TExpression.AllocateTempVar:string;
begin
result:=shader.AllocateTempVar;
end;

procedure TExpression.FreeTempVar(const s:string);
begin
shader.FreeTempVar(s);
end;

function TExpression.WriteBinaryOperator(const target:string=''):string;
var
 p1,p2,op,temp,s:string;
begin
p1:=subnodes[0].WriteExpression;
p2:=subnodes[1].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
result:='';
case info.token of
 tkadd,tksub,tkmult,tkor,tkand,tkgequal,tkless:
 begin
  case info.token of
   tkadd:op:='ADD';
   tksub:op:='SUB';
   tkmult:op:='MUL';
   tkor:op:='MAX';
   tkand:op:='MIN';
   tkgequal:op:='SGE';
   tkless:op:='SLT';
  end;
  if target='' then
  begin
   result:=AllocateTempVar;
   WriteCmd(op+sat,result,p1,p2);
  end
  else
  begin
   WriteCmd(op+sat,target,p1,p2);
   result:=target;
  end;
 end;
 tkgreater:
 begin
  if target='' then
  begin
   result:=AllocateTempVar;
   WriteCmd('SLT'+sat,result,NegateStr(p1),NegateStr(p2));
  end
  else
  begin
   WriteCmd('SLT'+sat,target,NegateStr(p1),NegateStr(p2));
   result:=target;
  end;
 end;
 tklequal:
 begin
  if target='' then
  begin
   result:=AllocateTempVar;
   WriteCmd('SGE'+sat,result,NegateStr(p1),NegateStr(p2));
  end
  else
  begin
   WriteCmd('SGE'+sat,target,NegateStr(p1),NegateStr(p2));
   result:=target;
  end;
 end;
 tkequal:
 begin
   result:=AllocateTempVar;
   WriteCmd('SUB',result,p1,p2);
   WriteCmd('ABS',result,result,'');
   if target<>'' then
   begin
    freetempvar(result);
    result:=target;
   end;
   WriteCmd('SLT'+sat,result,p1,'0.001');
   end;
 tknotequal,tkxor:
 begin
   result:=AllocateTempVar;
   WriteCmd('SUB',result,p1,p2);
   WriteCmd('ABS',result,result,'');
     if target<>'' then
   begin
    freetempvar(result);
    result:=target;
   end;
   WriteCmd('SGE'+sat,result,p1,'0.001');
   end;
 tkdiv:
 begin
   result:=AllocateTempVar;
   s:=GetSwizzleMask(p2);
   if length(s) in [0,4] then
   begin
   if s='' then
   begin
    s:='xyzw';
    p2:=p2+'.';
   end
   else
    p2:=copy(p2,1,length(p2)-length(s));
   WriteCmd('RCP',result+'.x',p2+s[1],'');
   WriteCmd('RCP',result+'.y',p2+s[2],'');
   WriteCmd('RCP',result+'.z',p2+s[3],'');
   WriteCmd('RCP',result+'.w',p2+s[4],'');
   end
   else
   begin
   p2:=ChangeScalarSwizzle(p2);
    WriteCmd('RCP',result,p2,'');
   end;
   if target<>'' then
   begin
    temp:=result;
    result:=target;
    WriteCmd('MUL'+sat,result,temp,p1);
    FreeTempVar(temp);
   end
   else
    WriteCmd('MUL'+sat,result,result,p1);
  end;
end;
end;

function TExpression.WriteUnaryOperator(const target:string=''):string;
var
 p1:string;
begin
case info.token of
 tksub:
 begin
 if target='' then
 begin
  if satfunc then
  begin
   p1:=subnodes[0].WriteExpression;
   FreeTempVar(p1);
   result:=AllocateTempVar;
   WriteCmd('MOV'+sat,result,negatestr(p1),'');
  end
  else
   result:=negatestr(subnodes[0].WriteExpression(target));
 end
 else
 begin
  p1:=subnodes[0].WriteExpression;
  FreeTempVar(p1);
  WriteCmd('MOV'+sat,target,negatestr(p1),'');
  result:=target;
 end;
 end;
 tkadd:
 begin
 if target='' then
 begin
  if satfunc then
  begin
   p1:=subnodes[0].WriteExpression;
   FreeTempVar(p1);
   result:=AllocateTempVar;
   WriteCmd('MOV'+sat,result,p1,'');
  end
  else
   result:=subnodes[0].WriteExpression(target);
 end
 else
 begin
  p1:=subnodes[0].WriteExpression;
  FreeTempVar(p1);
  WriteCmd('MOV'+sat,target,p1,'');
  result:=target;
 end;
 end;
 tknot:
 begin
  p1:=subnodes[0].WriteExpressionIdent;
  FreeTempVar(p1);
  if target='' then
  begin
   result:=AllocateTempVar;
   WriteCmd('SUB'+sat,result,'1',p1);
  end
  else
  begin
   WriteCmd('SUB'+sat,target,'1',p1);
   result:=target;
  end;
 end;
end;
end;

function TExpression.WriteOperator(const target:string=''):string;
begin
if unary then
 result:=WriteUnaryOperator(target)
else
 result:=WriteBinaryOperator(target);
end;

function TExpression.WriteExpressionIdent(const target:string=''):string;
begin
 CheckIdent;
 if target='' then
 begin
 if satfunc then
 begin
  result:=AllocateTempVar;
  WriteCmd('MOV'+sat,result,ident,'');
 end
 else
  result:=ident;
 end
 else
 begin
  result:=target;
  WriteCmd('MOV'+sat,target,ident,'');

 end;
end;

function TExpression.WriteExpressionConst(const target:string=''):string;
var
 f:single;
begin
 if target='' then
 begin
  if satfunc then
  begin
   f:=strtofloat(info.tokenstring);
   if f<0 then f:=0;
   if f>1 then f:=1;
   info.tokenstring:=floattostr(f);
  end;
  result:='{'+info.tokenstring+','+info.tokenstring+','+info.tokenstring+','+info.tokenstring+'}';
 end
 else
 begin
  result:=target;
  WriteCmd('MOV'+sat,target,'{'+info.tokenstring+','+info.tokenstring+','+info.tokenstring+','+info.tokenstring+'}','');
 end;

end;


function TExpression.WriteExpressionFunc_OpCode1(const opcode:string;const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(1);
p1:=paramlist[0].WriteExpression;
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd(opcode+sat,result,p1,'');
end
else
begin
 WriteCmd(opcode+sat,target,p1,'');
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_OpCode1s(const opcode:string;const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(1);
p1:=paramlist[0].WriteExpression;
FreeTempVar(p1);
p1:=ChangeScalarSwizzle(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd(opcode+sat,result,p1,'');
end
else
begin
 WriteCmd(opcode+sat,target,p1,'');
 result:=target;
end;
end;


function TExpression.WriteExpressionFunc_OpCode2(const opcode:string;const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(2);
p1:=paramlist[0].WriteExpression;
p2:=paramlist[1].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd(opcode+sat,result,p1,p2);
end
else
begin
 WriteCmd(opcode+sat,target,p1,p2);
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_OpCode2s(const opcode:string;const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(2);
p1:=paramlist[0].WriteExpression;
p2:=paramlist[1].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
p1:=ChangeScalarSwizzle(p1);
p2:=ChangeScalarSwizzle(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd(opcode+sat,result,p1,p2);
end
else
begin
 WriteCmd(opcode+sat,target,p1,p2);
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_rcp(const target:string=''):string;
var
 p1,sw:string;
begin
CheckParamCount(1);
p1:=paramlist[0].WriteExpression;
FreeTempVar(p1);
sw:=GetSwizzleMask(p1);
if target='' then
 result:=AllocateTempVar
else
 result:=target;
if length(sw) in [0,4] then
begin
 if sw='' then
 begin
  sw:='xyzw';
  p1:=p1+'.';
 end
 else
  p1:=copy(p1,1,length(p1)-length(sw));
 WriteCmd('RCP',result+'.x',p1+sw[1],'');
 WriteCmd('RCP',result+'.y',p1+sw[2],'');
 WriteCmd('RCP',result+'.z',p1+sw[3],'');
 WriteCmd('RCP',result+'.w',p1+sw[4],'');
end
else
begin
 p1:=ChangeScalarSwizzle(p1);
 WriteCmd('RCP',result,p1,'');
end;
end;


function TExpression.WriteExpressionFunc_Normalize(const target:string=''):string;
var
 p1,sw:string;
begin
CheckParamCount(1);
p1:=paramlist[0].WriteExpression;
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result+'.w',p1,p1);
 WriteCmd('RSQ',result+'.w',result+'.w','');
 WriteCmd('MUL'+sat,result+'.xyz',result+'.w',p1);
end
else
begin
 sw:=CombineSwizzle(target,'xyz');
 if sw<>'' then
 begin
  result:=AllocateTempVar;
  WriteCmd('DP3',result+'.w',p1,p1);
  WriteCmd('RSQ',result+'.w',result+'.w','');
  WriteCmd('MUL'+sat,sw,result+'.w',p1);
  FreeTempVar(result);
 end;
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_Dist(const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(2);
p1:=paramlist[0].WriteExpression;
p2:=paramlist[1].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('DP3',result+'.w',result,result);
 WriteCmd('RSQ',result+'.w,',result+'.w','');
 WriteCmd('RCP'+sat,result,result+'.w','');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('DP3',result+'.w',result,result);
 WriteCmd('RSQ',result+'.w',result+'.w','');
 WriteCmd('RCP'+sat,target,result+'.w','');
 FreeTempVar(result);
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_InvDist(const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(2);
p1:=paramlist[0].WriteExpression;
p2:=paramlist[1].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('DP3',result+'.w',result,result);
 WriteCmd('RSQ'+sat,result,result+'.w','');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('DP3',result+'.w',result,result);
 WriteCmd('RSQ'+sat,target,result+'.w','');
 FreeTempVar(result);
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_SqrDist(const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(2);
p1:=paramlist[0].WriteExpression;
p2:=paramlist[1].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('DP3'+sat,result,result,result);
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('DP3'+sat,target,result,result);
 FreeTempVar(result);
 result:=target;
end;
end;


function TExpression.WriteExpressionFunc_Length(const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(1);
p1:=paramlist[0].WriteExpression;
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result+'.w',p1,p1);
 WriteCmd('RSQ',result+'.w',result+'.w','');
 WriteCmd('RCP'+sat,result,result+'.w','');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result+'.w',p1,p1);
 WriteCmd('RSQ',result+'.w',result+'.w','');
 WriteCmd('RCP'+sat,target,result+'.w','');
 FreeTempVar(result);
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_SqrLength(const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(1);
p1:=paramlist[0].WriteExpression;
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP3'+sat,result,p1,p1);
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('DP3'+sat,target,p1,p1);
 FreeTempVar(result);
 result:=target;
end;
end;


function TExpression.WriteExpressionFunc_InvLength(const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(1);
p1:=paramlist[0].WriteExpression;
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result+'.w',p1,p1);
 WriteCmd('RSQ'+sat,result,result+'.w','');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result+'.w',p1,p1);
 WriteCmd('RSQ'+sat,target,result+'.w','');
 FreeTempVar(result);
 result:=target;

end;
end;


function TExpression.WriteExpressionFunc_sqr(const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(1);
p1:=paramlist[0].WriteExpression;
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('MUL'+sat,result,p1,p1);
end
else
begin
 WriteCmd('MUL'+sat,target,p1,p1);
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_sqrt(const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(1);
p1:=paramlist[0].WriteExpression;
FreeTempVar(p1);
p1:=ChangeScalarSwizzle(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('RSQ',result+'.w',p1,'');
 WriteCmd('RCP'+sat,result,result+'.w','');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('RSQ',result+'.w',p1,'');
 WriteCmd('RCP'+sat,target,result+'.w','');

 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_mad(const target:string=''):string;
var
 p1,p2,p3:string;
begin
CheckParamCount(3);
p1:=paramlist[0].WriteExpression;
p2:=paramlist[1].WriteExpression;
p3:=paramlist[2].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
FreeTempVar(p3);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('MAD'+sat,result,p1,p2,p3);
end
else
begin
 WriteCmd('MAD'+sat,target,p1,p2,p3);
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_MatrixMult3(const target:string=''):string;
var
 p1,p2,sw:string;
 row1,row2,row3:string;
begin
CheckParamCount(2);
if paramlist[0].info.token<>tkident then
 raise ESyntaxError.Create(TokenInfoStr(paramlist[0].info)+' Identifier expected ');

p2:=paramlist[1].WriteExpression;
FreeTempVar(p2);

if sametext(paramlist[0].ident,'matrix') then
begin
 if length(paramlist[0].paramlist)<>3 then
  raise ESyntaxError.Create(TokenInfoStr(paramlist[0].info)+' Invalid parameter count ');
row1:=paramlist[0].paramlist[0].WriteExpression;
row2:=paramlist[0].paramlist[1].WriteExpression;
row3:=paramlist[0].paramlist[2].WriteExpression;
FreeTempVar(row1);
FreeTempVar(row2);
FreeTempVar(row3);
end
else
begin
p1:=shader.FindValidMatrix(lowercase(paramlist[0].ident));
if p1='' then
 raise ESyntaxError.Create(TokenInfoStr(paramlist[0].info)+' Matrix expected ');
row1:=p1+'.row[0]';
row2:=p1+'.row[1]';
row3:=p1+'.row[2]';
end;

if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP3'+sat,result+'.x',row1,p2);
 WriteCmd('DP3'+sat,result+'.y',row2,p2);
 WriteCmd('DP3'+sat,result+'.z',row3,p2);
end
else
begin
 sw:=CombineSwizzle(target,'x');
 if sw<>'' then WriteCmd('DP3'+sat,sw,row1,p2);
  sw:=CombineSwizzle(target,'y');
 if sw<>'' then WriteCmd('DP3'+sat,sw,row2,p2);
  sw:=CombineSwizzle(target,'z');
 if sw<>'' then WriteCmd('DP3'+sat,sw,row3,p2);
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_MatrixMult4(const target:string=''):string;
var
 p1,p2,sw:string;
 row1,row2,row3,row4:string;
begin
CheckParamCount(2);
if paramlist[0].info.token<>tkident then
 raise ESyntaxError.Create(TokenInfoStr(paramlist[0].info)+' Identifier expected ');

p2:=paramlist[1].WriteExpression;
FreeTempVar(p2);

if sametext(paramlist[0].ident,'matrix') then
begin
 if length(paramlist[0].paramlist)<>4 then
  raise ESyntaxError.Create(TokenInfoStr(paramlist[0].info)+' Invalid parameter count ');
row1:=paramlist[0].paramlist[0].WriteExpression;
row2:=paramlist[0].paramlist[1].WriteExpression;
row3:=paramlist[0].paramlist[2].WriteExpression;
row4:=paramlist[0].paramlist[3].WriteExpression;
FreeTempVar(row1);
FreeTempVar(row2);
FreeTempVar(row3);
FreeTempVar(row4);
end
else
begin
p1:=shader.FindValidMatrix(lowercase(paramlist[0].ident));
if p1='' then
 raise ESyntaxError.Create(TokenInfoStr(paramlist[0].info)+' Matrix expected ');
row1:=p1+'.row[0]';
row2:=p1+'.row[1]';
row3:=p1+'.row[2]';
row4:=p1+'.row[3]';
end;

if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP4'+sat,result+'.x',row1,p2);
 WriteCmd('DP4'+sat,result+'.y',row2,p2);
 WriteCmd('DP4'+sat,result+'.z',row3,p2);
 WriteCmd('DP4'+sat,result+'.w',row4,p2);
end
else
begin
 sw:=CombineSwizzle(target,'x');
 if sw<>'' then WriteCmd('DP4'+sat,sw,row1,p2);
  sw:=CombineSwizzle(target,'y');
 if sw<>'' then WriteCmd('DP4'+sat,sw,row2,p2);
  sw:=CombineSwizzle(target,'z');
 if sw<>'' then WriteCmd('DP4'+sat,sw,row3,p2);
  sw:=CombineSwizzle(target,'w');
 if sw<>'' then WriteCmd('DP4'+sat,sw,row4,p2);

 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_Vector(const target:string=''):string;
var
 p1,p2,p3,p4,s,sw:string;
 c,i:integer;
 bl:array[0..3] of boolean;
begin
CheckParamCount(4);
p1:=paramlist[0].WriteExpression;
p2:=paramlist[1].WriteExpression;
p3:=paramlist[2].WriteExpression;
p4:=paramlist[3].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
FreeTempVar(p3);
FreeTempVar(p4);
s:='{';
c:=0;
for i:=0 to 3 do
begin
 if i>0 then s:=s+',';
 if paramlist[i].info.token in [tkinteger,tkfloat] then
 begin
  bl[i]:=false;
  inc(c);
  s:=s+paramlist[i].info.tokenstring;
 end
 else
 begin
  s:=s+'0';
  bl[i]:=true;
 end;
end;
s:=s+'}';
if target='' then
begin
 if c=4 then
 begin
  result:=s;
 end
 else
 begin
  result:=AllocateTempVar;
 if c>0 then
  WriteCmd('MOV'+sat,result,s,'');
 if bl[0] then WriteCmd('MOV'+sat,result+'.x',p1,'');
 if bl[1] then WriteCmd('MOV'+sat,result+'.y',p2,'');
 if bl[2] then WriteCmd('MOV'+sat,result+'.z',p3,'');
 if bl[3] then WriteCmd('MOV'+sat,result+'.w',p4,'');
 end;
end
else
begin
 if c>0 then
  WriteCmd('MOV'+sat,target,s,'');
 sw:=CombineSwizzle(target,'x');
 if (sw<>'') and bl[0] then WriteCmd('MOV'+sat,sw,p1,'');
 sw:=CombineSwizzle(target,'y');
 if (sw<>'') and bl[1] then WriteCmd('MOV'+sat,sw,p2,'');
 sw:=CombineSwizzle(target,'z');
 if (sw<>'') and bl[2] then WriteCmd('MOV'+sat,sw,p3,'');
 sw:=CombineSwizzle(target,'w');
 if (sw<>'') and bl[3] then WriteCmd('MOV'+sat,sw,p4,'');
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_sat(const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(1);
p1:=paramlist[0].WriteExpression;
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('MIN',result,p1,'1');
 WriteCmd('MAX',result,result,'0');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('MIN',result,p1,'1');
 WriteCmd('MAX',target,result,'0');
 FreeTempVar(result);
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_clamp(const target:string=''):string;
var
 p1,p2,p3:string;
begin
CheckParamCount(3);
p1:=paramlist[0].WriteExpression;
p2:=paramlist[1].WriteExpression;
p3:=paramlist[2].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
FreeTempVar(p3);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('MIN',result,p1,p3);
 WriteCmd('MAX'+sat,result,result,p2);
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('MIN',result,p1,p3);
 WriteCmd('MAX'+sat,target,result,p2);
 FreeTempVar(result);
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_reflect(const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(2);
p1:=paramlist[0].WriteExpression;
p2:=paramlist[1].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result,p1,p2);
 WriteCmd('MUL',result,result,p2);
 WriteCmd('MAD'+sat,result,result,'{-2,-2,-2,-2}',p1);
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result,p1,p2);
 WriteCmd('MUL',result,result,p2);
 WriteCmd('MAD'+sat,target,result,'{-2,-2,-2,-2}',p1);
 FreeTempVar(result);
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_lightatten(const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(2);
p1:=paramlist[0].WriteExpression;
p2:=paramlist[1].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('MUL',result,p1,p2);
 WriteCmd('DP3',result,result,result);
 WriteCmd('ADD'+sat,result,'{1,1,1,1}',NegateStr(result));
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('MUL',result,p1,p2);
 WriteCmd('DP3',result,result,result);
 WriteCmd('ADD'+sat,target,'{1,1,1,1}',NegateStr(result));
 FreeTempVar(result);
 result:=target;
end;
end;

function TExpression.WriteExpressionFunc_Decl(func:TFunctionInfo;const target:string=''):string;
var
 i:integer;
 sf:string;
 ws:integer;
begin
if func.called then
 raise ESyntaxError.Create(TokenInfoStr(info)+' Recursive function call not allowed');
CheckParamCount(length(func.paramlist));
for i:=0 to length(paramlist)-1 do
 paramlist[i].WriteExpression(func.name+'$'+func.paramlist[i]);
sf:=shader.currentfunc;
shader.currentfunc:=func.name+'$';
ws:=shader.withlistart;
shader.withlistart:=shader.withlist.Count;
func.codeblock.Write;
shader.currentfunc:=sf;
shader.withlistart:=ws;
if target='' then
 result:=func.name+'$Result'
else
begin
 result:=target;
 WriteCmd('MOV',target,func.name+'$Result','');
end;
end;

function TExpression.WriteExpressionFunc(const target:string=''):string;
var
 fname:string;
 fd:TFunctionInfo;
begin
result:='';
fname:=lowercase(ident);
fd:=shader.FindFunction(fname);
if fd<>nil then result:=WriteExpressionFunc_Decl(fd,target) else
if fname='min' then result:=WriteExpressionFunc_OpCode2('MIN',target) else
if fname='max' then result:=WriteExpressionFunc_OpCode2('MAX',target) else
if fname='abs' then result:=WriteExpressionFunc_OpCode1('ABS',target) else
if fname='cross' then result:=WriteExpressionFunc_OpCode2('XPD',target) else
if (fname='dot') or (fname='dot3') then result:=WriteExpressionFunc_OpCode2('DP3',target) else
if fname='dot4' then result:=WriteExpressionFunc_OpCode2('DP4',target) else
if fname='doth' then result:=WriteExpressionFunc_OpCode2('DPH',target) else
if fname='floor' then result:=WriteExpressionFunc_OpCode1('FLR',target) else
if fname='frac' then result:=WriteExpressionFunc_OpCode1('FRC',target) else
if fname='exp2' then result:=WriteExpressionFunc_OpCode1s('EX2',target) else
if fname='log2' then result:=WriteExpressionFunc_OpCode1s('LG2',target) else
if fname='power' then result:=WriteExpressionFunc_OpCode2s('POW',target) else
if fname='rcp' then result:=WriteExpressionFunc_rcp(target) else
if (fname='rsq') or (fname='rsqrt') then result:=WriteExpressionFunc_OpCode1s('RSQ',target) else
if fname='normalize' then result:=WriteExpressionFunc_Normalize(target) else
if fname='length' then result:=WriteExpressionFunc_Length(target) else
if fname='sqrlength' then result:=WriteExpressionFunc_SqrLength(target) else
if fname='invlength' then result:=WriteExpressionFunc_InvLength(target) else
if fname='dist' then result:=WriteExpressionFunc_Dist(target) else
if fname='sqrdist' then result:=WriteExpressionFunc_SqrDist(target) else
if fname='invdist' then result:=WriteExpressionFunc_InvDist(target) else
if fname='sqr' then result:=WriteExpressionFunc_sqr(target) else
if fname='sqrt' then result:=WriteExpressionFunc_sqrt(target) else
if fname='mad' then result:=WriteExpressionFunc_mad(target) else
if fname='matrixmult3' then result:=WriteExpressionFunc_MatrixMult3(target) else
if fname='matrixmult4' then result:=WriteExpressionFunc_MatrixMult4(target) else
if fname='vector' then result:=WriteExpressionFunc_Vector(target) else
if fname='sat' then result:=WriteExpressionFunc_sat(target) else
if fname='clamp' then result:=WriteExpressionFunc_clamp(target) else
if fname='lightatten' then result:=WriteExpressionFunc_lightatten(target) else
if fname='reflect' then result:=WriteExpressionFunc_reflect(target)
 else result:=shader.WriteExpressionFunction(self,target);
if result='' then
 raise ESyntaxError.Create(TokenInfoStr(info)+'Undeclared identifier: "'+ident+'"');
end;


function TExpression.WriteExpression(target:string=''):string;
begin
case info.token of
 tkident:
  begin
   if paramlist=nil then
     result:=WriteExpressionIdent(target)
   else
    result:=WriteExpressionFunc(target);
  end;
 tkinteger,tkfloat:result:=WriteExpressionConst(target);
 tkadd,tksub,tkmult,tkdiv,tkand,
 tkor,tkxor,tkequal,tknotequal,tknot,
 tkgequal,tkgreater,tkless,tklequal:
  begin
   result:=WriteOperator(target);
  end;
end;
end;


constructor TExpression.create(ps:TPascalShader);
begin
inherited create;
shader:=ps;
shader.nodelist.add(self);
info:=shader.ReadTokenInfo;
end;

// TStatement

procedure TStatement.WriteCmd(const cmd,p1,p2,p3:string;const p4:string='');
begin
shader.WriteCmd(cmd,p1,p2,p3,p4);
end;

function TStatement.AllocateTempVar:string;
begin
result:=shader.AllocateTempVar;
end;

procedure TStatement.FreeTempVar(const s:string);
begin
shader.FreeTempVar(s);
end;


procedure TStatement.WriteAssignment;
var
 p1,p2:string;
begin
exp1.CheckIdent;
if not (exp1.symbolinfo.scope in [scvar,scoutput]) then
 raise ESyntaxError.Create(TokenInfoStr(info)+'Left side cannot be assigned to');
exp2:=exp2.Optimize;

if shader.currentcondition='' then
begin
 p1:=exp2.WriteExpression(exp1.WriteExpression);
 FreeTempVar(p1);  // wirklich ntig?
end
else
begin // Ausdruck: x=x+...
 if (exp2.info.token=tkadd) and (exp2.subnodes[0].info.token=tkident) and sametext(exp2.subnodes[0].ident,exp1.ident) then
 begin
  p1:=exp1.WriteExpression;
  p2:=exp2.subnodes[1].WriteExpression;
  WriteCmd('MAD'+exp2.sat,p1,p2,shader.currentcondition,p1);
  end
 else // Ausdruck: x=...-x
 if (exp2.info.token=tkadd) and (exp2.subnodes[1].info.token=tkident) and sametext(exp2.subnodes[1].ident,exp1.ident) then
 begin
  p1:=exp1.WriteExpression;
  p2:=exp2.subnodes[0].WriteExpression;
  WriteCmd('MAD'+exp2.sat,p1,p2,shader.currentcondition,p1);
  end
 else // Ausdruck: x=x-...
 if (exp2.info.token=tksub) and (exp2.subnodes[0].info.token=tkident) and sametext(exp2.subnodes[0].ident,exp1.ident) then
 begin
  p1:=exp1.WriteExpression;
  p2:=exp2.subnodes[1].WriteExpression;
  WriteCmd('MAD'+exp2.sat,p1,p2,NegateStr(shader.currentcondition),p1);
  end
 else  // Ausdruck: x=...-x
 if (exp2.info.token=tksub) and (exp2.subnodes[1].info.token=tkident) and sametext(exp2.subnodes[1].ident,exp1.ident) then
 begin
  p1:=exp1.WriteExpression;
  p2:=exp2.subnodes[0].WriteExpression;
  WriteCmd('MAD'+exp2.sat,p1,p2,NegateStr(shader.currentcondition),p1);
  end
 else
  begin
   if shader.CMP_Cmd then
   begin
    p1:=exp1.WriteExpression;
    p2:=exp2.WriteExpression;
    WriteCmd('CMP',p1,shader.currentcondition_cmp,p2,p1);
    FreeTempVar(p2);
   end
   else
   begin
    p1:=exp1.WriteExpression;
    WriteCmd('MUL',p1,p1,shader.currentconditioninv);
    p2:=exp2.WriteExpression;
    WriteCmd('MAD',p1,p2,shader.currentcondition,p1);
    FreeTempVar(p2);
   end;
  end;
 end;
end;

procedure TStatement.WriteCodeBlock;
var
 i:integer;
begin
for i:=0 to length(statements)-1 do
 statements[i].Write;
end;

procedure TStatement.WriteWithStatement;
var
 s:string;
begin
s:=shader.FindValidWithDecl(withdecl);
if s='' then
 raise ESyntaxError.Create(TokenInfoStr(info)+' Undeclared Identifier: "'+withdecl+'"');
shader.withlist.Add(s);
statements[0].write;
shader.withlist.Delete(shader.withlist.count-1);
end;

procedure TStatement.WriteIfStatement;
var
 cond,sgc,sgcinv,s:string;
begin
cond:=exp1.WriteExpression;
sgc:=shader.currentcondition;
sgcinv:=shader.currentconditioninv;


shader.currentconditioninv:=AllocateTempVar;
if sgc<>'' then
begin
 shader.currentcondition:=AllocateTempVar;
 WriteCmd('MIN',shader.currentcondition,sgc,cond);
end
else
 shader.currentcondition:=cond;
WriteCmd('SUB',shader.currentconditioninv,'1',shader.currentcondition);

if shader.CMP_Cmd then
begin
shader.currentcondition_cmp:=AllocateTempVar;
WriteCmd('SUB',shader.currentcondition_cmp,'{0.5,0.5,0.5,0.5}',shader.currentcondition);
end;

statements[0].write;
if shader.CMP_Cmd then
 shader.currentcondition_cmp:=NegateStr(shader.currentcondition_cmp);
if length(statements)>1 then
begin
s:=shader.currentcondition;
shader.currentconditioninv:=shader.currentcondition;
shader.currentcondition:=s;
statements[1].write;
end;

if shader.CMP_Cmd then
begin
 shader.currentcondition_cmp:=NegateStr(shader.currentcondition_cmp);
 FreeTempVar(shader.currentcondition_cmp);
end;

shader.FreeTempVar(shader.currentcondition);
shader.FreeTempVar(shader.currentconditioninv);

shader.currentcondition:=sgc;
shader.currentconditioninv:=sgc;

end;

procedure TStatement.WriteDiscard;
begin
if shader.currentcondition='' then
 WriteCmd('KIL','','{-1,-1,-1,-1}','','')
else
 WriteCmd('KIL','',shader.currentcondition_cmp,'','');
end;


procedure TStatement.Write;
begin
case _type of
 stassignment:WriteAssignment;
 stblock:WriteCodeBlock;
 stwith:WriteWithStatement;
 stif:WriteIfStatement;
 stdiscard:WriteDiscard;
end;
end;


constructor TStatement.create(ps:TPascalShader);
begin
inherited create;
shader:=ps;
shader.nodelist.Add(self);
end;

// TARBCommand

function TARBCommand.IsVarRead(const v:string):boolean;
begin
result:=(Sametext(ExtractVarname(p2),v) or
        Sametext(ExtractVarname(p3),v) or Sametext(ExtractVarname(p4),v)) and (copy(v,1,1)<>'{') and (copy(v,1,2)<>'-{');
end;

function TARBCommand.IsVarOverWritten(const v:string):boolean;
var
 sw:string;
begin
sw:=GetSwizzleMask(p1);
result:=(v<>'') and Sametext(ExtractVarname(p1),v) and ((sw='') or Sametext(sw,'xyzw'))and (copy(v,1,1)<>'{') and (copy(v,1,2)<>'-{');
end;


function TARBCommand.IsVarWritten(const v:string):boolean;
var
 sw:string;
begin
sw:=GetSwizzleMask(p1);
result:=(v<>'') and Sametext(ExtractVarname(p1),v) and (copy(v,1,1)<>'{')and (copy(v,1,2)<>'-{');
end;

function TARBCommand.IsVarCompWritten(const v,comp:string):boolean;
var
 sw:string;
begin
sw:=GetSwizzleMask(p1);
result:=(v<>'') and Sametext(ExtractVarname(p1),v) and ((pos(comp,sw)<>0) or (sw='')) and (copy(v,1,1)<>'{') and (copy(v,1,2)<>'-{');
end;


constructor TARBCommand.create(ps:TPascalShader);
begin
inherited create;
shader:=ps;
end;

// TPascalShader

function TPascalShader.ReadTokenInfo:TTokenInfo;
begin
result.token:=sc.TokenType;
result.SourcePos:=sc.SourcePos;
result.SourceLine:=sc.SourceLine;
result.tokenstring:=sc.TokenString;
end;

procedure TPascalShader.ParseVarDecl(fd:TFunctionInfo);
var
 si:TSymbolInfo;
begin
repeat
 if (FindSymbol(sc.TokenString)<>nil) or (FindFunction(sc.tokenstring)<>nil) or (reservedwordlist.IndexOf(sc.TokenString)<>-1) then
  raise ESyntaxError.Create(TokenInfoStr(ReadTokenInfo)+'Identifier redeclared: "'+sc.TokenString+'"');
 si:=TSymbolInfo.Create;
 if fd=nil then
  si.name:=sc.TokenString
 else
  si.name:=fd.name+'$'+sc.TokenString;
 si.scope:=scvar;
 symbollist.Add(si);
 sc.AcceptToken(tkident);
sc.CheckToken([tkcomma,tksemicolon]);
if sc.tokentype=tkcomma then sc.AcceptToken(tkcomma);
until sc.TokenType=tksemicolon;
sc.AcceptToken(tksemicolon);
end;

procedure TPascalShader.ParseVarBlock(fd:TFunctionInfo);
begin
sc.AcceptToken(tkvar);
if fd=nil then
 sc.CheckToken([tkident,tkbegin,tkparam,tkvar,tkvarying])
else
 sc.CheckToken([tkident,tkbegin,tkparam,tkvar,tkvarying,tkfunction]);
while sc.TokenType=tkident do
begin
 case sc.TokenType of
  tkIdent:
   begin
    ParseVarDecl(fd);
   end;
 end;
end;

end;

procedure TPascalShader.ParseParamDecl;
var
 si,si2:TSymbolInfo;
 i,c:integer;
begin
repeat
 if (FindSymbol(sc.TokenString)<>nil) or (FindFunction(sc.tokenstring)<>nil) or (reservedwordlist.IndexOf(sc.TokenString)<>-1) then
  raise ESyntaxError.Create(TokenInfoStr(ReadTokenInfo)+'Identifier redeclared: "'+sc.TokenString+'"');;
 si:=TSymbolInfo.Create;
 si.name:=sc.TokenString;
 si.scope:=scparam;
 sc.AcceptToken(tkident);
 sc.CheckToken([tkcomma,tksemicolon,tkarrayopen]);
 if sc.tokentype=tkarrayopen then
 begin
  sc.AcceptToken(tkarrayopen);
  sc.CheckToken([tkinteger]);
  c:=strtoint(sc.tokenstring);
  sc.AcceptToken(tkinteger);
  sc.AcceptToken(tkarrayclose);
  for i:=0 to c-1 do
  begin
   si2:=TSymbolInfo.Create;
   if i=0 then
    si2.arraysize:=c
   else
    si2.arraysize:=-1;
   si2.offset:=paramsize+i;
   si2.name:=si.name+'['+inttostr(i)+']';
   si2.scope:=si.scope;
   symbollist.add(si2);
  end;
  inc(paramsize,c);
  si.free;
 end
 else
  begin
  symbollist.add(si);
  si.offset:=paramsize;
  inc(paramsize);
  end;
sc.CheckToken([tkcomma,tksemicolon]);
if sc.tokentype=tkcomma then sc.AcceptToken(tkcomma);
until sc.TokenType=tksemicolon;
sc.AcceptToken(tksemicolon);
end;

procedure TPascalShader.ParseParamBlock;
begin
sc.AcceptToken(tkparam);
sc.CheckToken([tkident,tkbegin,tkvar,tkparam,tkvarying,tkfunction]);
while sc.TokenType=tkident do
begin
 case sc.TokenType of
  tkIdent:
   begin
    ParseParamDecl;
   end;
 end;
end;

end;


procedure TPascalShader.ParseVaryingDecl;
var
 si:TSymbolInfo;
begin
repeat
 if (FindSymbol(sc.TokenString)<>nil) or (FindFunction(sc.tokenstring)<>nil) or (reservedwordlist.IndexOf(sc.TokenString)<>-1) then
  raise ESyntaxError.Create(TokenInfoStr(ReadTokenInfo)+'Identifier redeclared: "'+sc.TokenString+'"');
 si:=TSymbolInfo.Create;
 si.name:=sc.TokenString;
 si.scope:=VaryingType;
 si.varying:=true;
 symbollist.Add(si);
 sc.AcceptToken(tkident);
sc.CheckToken([tkcomma,tksemicolon]);
if sc.tokentype=tkcomma then sc.AcceptToken(tkcomma);
until sc.TokenType=tksemicolon;
sc.AcceptToken(tksemicolon);
end;

procedure TPascalShader.ParseVaryingBlock;
begin
sc.AcceptToken(tkvarying);
sc.CheckToken([tkident,tkbegin,tkparam,tkvar,tkvarying,tkfunction]);
while sc.TokenType=tkident do
begin
 case sc.TokenType of
  tkIdent:
   begin
    ParseVaryingDecl;
   end;
 end;
end;

end;

procedure TPascalShader.ParseFuncDecl;
var
 fd:TFunctionInfo;
 l:integer;
 si:TSymbolInfo;
begin
fd:=TFunctionInfo.Create;
funclist.Add(fd);

sc.AcceptToken(tkfunction);
sc.CheckToken([tkident]);
if (FindFunction(sc.TokenString)<>nil) or (FindSymbol(sc.TokenString)<>nil) or (reservedwordlist.IndexOf(sc.TokenString)<>-1)  then
 raise ESyntaxError.Create(TokenInfoStr(ReadTokenInfo)+' Identfier redeclared: "'+sc.TokenString+'"');
fd.name:=sc.TokenString;

si:=TSymbolInfo.Create;
si.name:=fd.name+'$Result';
si.scope:=scvar;
symbollist.Add(si);

sc.AcceptToken(tkident);
sc.CheckToken([tkopenbracket,tksemicolon]);
if sc.TokenType=tkopenbracket then
begin
 sc.AcceptToken(tkopenbracket);
 sc.CheckToken([tkident,tkclosebracket]);
 while sc.TokenType<>tkclosebracket do
 begin
  l:=length(fd.paramlist);
  setlength(fd.paramlist,l+1);
  fd.paramlist[l]:=sc.TokenString;
  si:=TSymbolInfo.Create;
  si.name:=fd.name+'$'+sc.TokenString;
  si.scope:=scvar;
  symbollist.Add(si);
  sc.AcceptToken(tkident);
  sc.CheckToken([tkcomma,tkclosebracket]);
  if sc.TokenType=tkcomma then
  begin
   sc.AcceptToken(tkcomma);
   sc.CheckToken([tkident]);
  end;
 end;
 sc.AcceptToken(tkclosebracket);
end;
sc.AcceptToken(tksemicolon);

sc.CheckToken([tkvar,tkbegin]);
while sc.TokenType<>tkbegin do
begin
 ParseVarBlock(fd);
sc.CheckToken([tkvar,tkbegin]);
end;

fd.codeblock:=ParseCodeBlock;
sc.AcceptToken(tksemicolon);
end;

function TPascalShader.ParseExpression_r:TExpression;
const
 endtoken=[tkcomma,tksemicolon,tkassignment,tkend,tkelse,tkthen];
var
 e:TExpression;
 preoperator:boolean;
 opstack:TStack;
 l:integer;
 b:boolean;
begin
sc.CheckToken([tkident,tkinteger,tkfloat,tkopenbracket]+unaryop);
preoperator:=true;
opstack:=TStack.Create;
repeat
case sc.TokenType of
 tkident:
  begin
   e:=TExpression.create(self);
   repeat
   b:=true;
   e.ident:=e.ident+sc.TokenString;
   sc.AcceptToken(tkident);
   if sc.TokenType=tkarrayopen then
   begin
    e.ident:=e.ident+sc.TokenString;
    sc.AcceptToken(tkarrayopen);
    e.ident:=e.ident+sc.TokenString;
    sc.AcceptToken(tkinteger);
    e.ident:=e.ident+sc.TokenString;
    sc.AcceptToken(tkarrayclose);
   end;
   if sc.TokenType=tkpoint then
   begin
   e.ident:=e.ident+sc.TokenString;
    sc.AcceptToken(tkpoint);
    sc.CheckToken([tkident]);
    b:=false;
   end;
   until b;
   if sc.TokenType=tkopenbracket then
   begin
    sc.AcceptToken(tkopenbracket);
    while sc.TokenType<>tkclosebracket do
    begin
     l:=length(e.paramlist);
     setlength(e.paramlist,l+1);
     e.paramlist[l]:=ParseExpression;
     sc.CheckToken([tkcomma,tkclosebracket]);
     if sc.TokenType=tkcomma then sc.AcceptToken(tkcomma);
    end;
    sc.AcceptToken(tkclosebracket);
   end;
   e.Push;
   sc.CheckToken(binaryop+endtoken+[tkarrayclose,tkclosebracket]);
   preoperator:=false;
  end;
 tkinteger,tkfloat:
  begin
   e:=TExpression.create(self);
   e.Push;
   sc.AcceptToken(sc.TokenType);
   sc.CheckToken(binaryop+endtoken+[tkclosebracket]);
   preoperator:=false;
  end;
  tkadd,tksub,tkmult,tkdiv,tkequal,
  tklequal,tkless,tkgequal,tkgreater,tknotequal,
  tkand,tkor,tkxor:
  begin
   e:=TExpression.Create(self);
   e.unary:=preoperator or (sc.TokenType in [tknot]);
   sc.AcceptToken(e.info.token);
   while (opstack.count>0) and TExpression(opstack.Peek).CompareOperator(e) do
   begin
    TExpression(opstack.Pop).AddOperator;
   end;
   opstack.push(e);
   if preoperator then
   begin
    sc.CheckToken(unaryop+[tkident,tkinteger,tkfloat,tkopenbracket]-[tkadd,tksub]);
    preoperator:=false;
   end
   else
   begin
    sc.CheckToken(unaryop+[tkident,tkinteger,tkfloat,tkopenbracket]);
    preoperator:=true;
   end;
  end;
 tkpoint:
 begin
  e:=TExpression.Create(self);
  e.unary:=false;
  sc.AcceptToken(e.info.token);
  while (opstack.count>0) and TExpression(opstack.Peek).CompareOperator(e) do
  begin
   TExpression(opstack.Pop).AddOperator;
  end;
  opstack.push(e);
  sc.CheckToken([tkident]);
  preoperator:=false;
 end;
 tkopenbracket:
 begin
  sc.AcceptToken(tkopenbracket);
  ParseExpression_r;
  sc.AcceptToken(tkclosebracket);
  preoperator:=false;
  sc.CheckToken(binaryop+endtoken+[tkclosebracket,tkderef]);
 end;
 tknot:
 begin
   e:=TExpression.Create(self);
   e.unary:=true;
   sc.AcceptToken(e.info.token);
   while (opstack.count>0) and TExpression(opstack.Peek).CompareOperator(e) do
   begin
    TExpression(opstack.Pop).AddOperator;
   end;
   opstack.push(e);
   sc.CheckToken(unaryop+[tkident,tkinteger,tkfloat,tkopenbracket]);
   preoperator:=true;
 end;
end;
until sc.TokenType in [tksemicolon,tkcomma,tkassignment,tkclosebracket,tkend,tkelse,tkthen];

while (opstack.count>0) do
begin
 TExpression(opstack.Pop).AddOperator;
end;
result:=exptree;
opstack.free;
end;

function TPascalShader.ParseExpression:TExpression;
var
 start:integer;
begin
start:=expstack.count;
result:=ParseExpression_r;
if expstack.count-start<>1 then
 raise ESyntaxError.Create(TokenInfoStr(result.info)+'Expression expected but "'+sc.TokenString+'" found');
expstack.pop;
end;

function TPascalShader.FindSymbol(const sname:string):TSymbolInfo;
var
 i,j:integer;
begin
result:=nil;
for j:=withlist.count-1 downto withlistart do
begin
 for i:=0 to symbollist.count-1 do
  if sametext(TSymbolInfo(symbollist[i]).name,sname) then
  begin
   result:=TSymbolInfo(symbollist[i]);
   break;
  end;
 if result<>nil then break;
end;
if result=nil then
for i:=0 to symbollist.count-1 do
 if sametext(TSymbolInfo(symbollist[i]).name,currentfunc+sname) then
 begin
  result:=TSymbolInfo(symbollist[i]);
  break;
 end;
end;

function TPascalShader.FindFunction(const fname:string):TFunctionInfo;
var
 i:integer;
begin
result:=nil;
for i:=0 to funclist.count-1 do
 if SameText(TFunctionInfo(funclist[i]).name,fname) then
 begin
  result:=TFunctionInfo(funclist[i]);
  break;
 end;
end;



function TPascalShader.FindValidMatrix(const s:string):string;
var
 i:integer;
begin
result:='';
 for i:=withlist.count-1 downto withlistart do
  if matrixlist.IndexOf(withlist[i]+'.'+s)<>-1 then
  begin
   result:=withlist[i]+'.'+s;
   break;
  end;
if (result='') and (matrixlist.IndexOf(s)<>-1) then
 result:=s;
end;

function TPascalShader.IsValidWithDecl(const s:string):boolean;
var
 i:integer;
begin
result:=false;
for i:=0 to matrixlist.count-1 do
 if sametext(s+'.',copy(matrixlist[i],1,length(s)+1)) then
 begin
  result:=true;
  break;
 end;
if not result then
for i:=0 to symbollist.count-1 do
 if sametext(s+'.',copy(TSymbolInfo(symbollist[i]).name,1,length(s)+1)) then
 begin
  result:=true;
  break;
 end;

end;

function TPascalShader.FindValidWithDecl(const s:string):string;
var
 i:integer;
begin
result:='';
 for i:=withlist.count-1 downto withlistart do
  if IsValidWithDecl(withlist[i]+'.'+s) then
  begin
   result:=withlist[i]+'.'+s;
   break;
  end;
if (result='') and IsValidWithDecl(s) then
 result:=s;

end;



function TPascalShader.ParseExpressionStatement:TStatement;
begin
result:=TStatement.Create(self);
result.exp1:=ParseExpression;
result.exp1.lvalue:=true;
sc.CheckToken([tkassignment]);
sc.AcceptToken(tkassignment);
result._type:=stassignment;
result.exp2:=ParseExpression;
end;

function TPascalShader.ParseCodeBlock:TStatement;
var
 l:integer;
begin
result:=TStatement.Create(self);
result._type:=stblock;
sc.AcceptToken(tkbegin);
sc.CheckToken([tkend,tkident,tkfloat,tkinteger,tkopenbracket,tkwith,tkdiscard,tkif]+expressionstart);
while sc.TokenType<>tkend do
begin
while sc.TokenType=tksemicolon do sc.AcceptToken(tksemicolon);
l:=length(result.statements);
setlength(result.statements,l+1);
result.statements[l]:=ParseStatement;
while sc.TokenType=tksemicolon do sc.AcceptToken(tksemicolon);
sc.CheckToken([tkend,tkident,tkfloat,tkinteger,tkopenbracket,tkwith,tkdiscard,tkif]+expressionstart);
end;
sc.AcceptToken(tkend);
end;

function TPascalShader.ParseWith:TStatement;
var
 b:boolean;
begin
result:=TStatement.Create(self);
result._type:=stwith;
result.info:=ReadTokenInfo;
sc.AcceptToken(tkwith);
sc.CheckToken([tkident]);
repeat
b:=true;
result.withdecl:=result.withdecl+sc.TokenString;
sc.AcceptToken(tkident);
if sc.TokenType=tkarrayopen then
begin
 result.withdecl:=result.withdecl+sc.TokenString;
 sc.AcceptToken(tkarrayopen);
 result.withdecl:=result.withdecl+sc.TokenString;
 sc.AcceptToken(tkinteger);
 result.withdecl:=result.withdecl+sc.TokenString;
 sc.AcceptToken(tkarrayclose);
end;
if sc.TokenType=tkpoint then
begin
 result.withdecl:=result.withdecl+sc.TokenString;
 sc.AcceptToken(tkpoint);
 sc.CheckToken([tkident]);
 b:=false;
end;
until b;
sc.AcceptToken(tkdo);
setlength(result.statements,1);
result.statements[0]:=ParseStatement;
end;

function TPascalShader.ParseDiscard:TStatement;
begin
result:=TStatement.Create(self);
result._type:=stdiscard;
result.info:=ReadTokenInfo;
if not AllowDiscard then
 raise ESyntaxError.Create(TokenInfoStr(ReadTokenInfo)+' Discard not allowed');
sc.AcceptToken(tkdiscard);
end;

function TPascalShader.Parseif:TStatement;
begin
result:=TStatement.Create(self);
result._type:=stif;
result.info:=ReadTokenInfo;
sc.AcceptToken(tkif);
Result.exp1:=ParseExpression;
sc.AcceptToken(tkthen);
setlength(result.statements,1);
result.statements[0]:=ParseStatement;
if sc.TokenType=tkelse then
begin
 sc.AcceptToken(tkelse);
 setlength(result.statements,2);
 result.statements[1]:=ParseStatement;
end;
end;


function TPascalShader.ParseStatement:TStatement;
var
 inf:TTokenInfo;
begin
sc.CheckToken([tkbegin,tkopenbracket,tkwith,tkdiscard,tkif,tkinteger,tkfloat,tkadd,tksub,tknot]+expressionstart);
inf:=ReadTokenInfo;
case sc.TokenType of
 tkbegin:result:=ParseCodeBlock;
 tkident,tkopenbracket,tkinteger,tkfloat,tkadd,tksub,tknot:
  begin
   result:=ParseExpressionStatement;
  end;
 tkwith:
  begin
  result:=ParseWith;
  end;
  tkdiscard:
   begin
   result:=ParseDiscard;
   end;
  tkif:
   begin
    result:=Parseif;
   end;
 else
  raise ESyntaxError.Create(TokenInfoStr(ReadTokenInfo)+'Statement expected');
end;
result.info:=inf;
end;


procedure TPascalShader.ParseProgram;
begin
sc.CheckToken([tkvar,tkparam,tkvarying,tkbegin,tkfunction]);
while sc.TokenType<>tkbegin do
begin
 case sc.TokenType of
  tkvar:ParseVarBlock(nil);
  tkparam:ParseParamBlock;
  tkvarying:ParseVaryingBlock;
  tkfunction:ParseFuncDecl;
 end;
end;
codeblock:=ParseCodeBlock;
end;

procedure TPascalShader.AddDefaultSymbol(const s:string;sc:TSymbolScope);
var
 si:TSymbolInfo;
begin
si:=TSymbolInfo.Create;
si.name:=s;
si.scope:=sc;
si._default:=true;
symbollist.Add(si)
end;

procedure TPascalShader.WriteLine(const s:string);
begin
fSource:=fSource+s+#13#10;
end;


procedure TPascalShader.WriteCmd(const cmd,p1,p2,p3:string;const p4:string='');
var
 l:integer;
begin
l:=length(cmdlist);
setlength(cmdlist,l+1);
cmdlist[l]:=TARBCommand.create(self);
cmdlist[l].cmd:=cmd;
cmdlist[l].p1:=p1;
cmdlist[l].p2:=p2;
cmdlist[l].p3:=p3;
cmdlist[l].p4:=p4;
end;

function TPascalShader.AllocateTempVar:string;
var
 si:TSymbolInfo;
 i:integer;
begin
if tempvarlist=nil then
begin
 i:=-1;
 repeat
inc(i);
  si:=FindSymbol('temp'+inttostr(i));
 until si=nil;
 si:=TSymbolInfo.Create;
 si.scope:=scvar;
 si.name:='temp'+inttostr(i);
 si.temp:=true;
 symbollist.add(si);
 result:=si.name;
end
else
begin
 result:=tempvarlist[length(tempvarlist)-1];
 setlength(tempvarlist,length(tempvarlist)-1);
end;
end;

procedure TPascalShader.FreeTempVar(const s:string);
var
 l:integer;
 si:TSymbolInfo;
begin
 si:=FindSymbol(s);
 if (si<>nil) and (si.temp) then
 begin
 l:=length(tempvarlist);
 setlength(tempvarlist,l+1);
 tempvarlist[l]:=si.name;
 end;
end;



function TPascalShader.WriteExpressionFunction(exp:TExpression;const target:string=''):string;
begin
result:='';
end;

procedure TPascalShader.WriteVarBlock;
var
 i:integer;
begin
for i:=0 to symbollist.count-1 do
if (TSymbolInfo(symbollist[i]).scope=scvar) and TSymbolInfo(symbollist[i]).used and not TSymbolInfo(symbollist[i])._default then
 WriteLine('TEMP '+TSymbolInfo(symbollist[i]).name+';');
end;

procedure TPascalShader.WriteVaryingBlock;
var
 c,i:integer;
begin
c:=0;
case VaryingType of
 scoutput:
  for i:=0 to symbollist.count-1 do
  if (TSymbolInfo(symbollist[i]).varying) then
  begin
   TSymbolInfo(symbollist[i]).offset:=c;
  if TSymbolInfo(symbollist[i]).used then
   WriteLine('OUTPUT '+TSymbolInfo(symbollist[i]).name+'=result.texcoord['+inttostr(c)+'];');
   inc(c);
  end;
 scattrib:
  for i:=0 to symbollist.count-1 do
  if (TSymbolInfo(symbollist[i]).varying) then
  begin
   TSymbolInfo(symbollist[i]).offset:=c;
   if TSymbolInfo(symbollist[i]).used then
    WriteLine('ATTRIB '+TSymbolInfo(symbollist[i]).name+'=fragment.texcoord['+inttostr(c)+'];');
   inc(c);
  end;
end;
end;


procedure TPascalShader.WriteParamBlock;
var
 i:integer;
begin
for i:=0 to symbollist.count-1 do
 with TSymbolInfo(symbollist[i]) do
 if (scope=scparam) and not _default and used then
 begin
  if arraysize=0 then
   WriteLine('PARAM '+name+'=program.local['+inttostr(offset)+'];')
  else
  if arraysize>0 then
   WriteLine('PARAM '+name+'=program.local['+inttostr(offset)+'..'+inttostr(offset+arraysize-1)+'];')
  end;
end;


procedure TPascalShader.WriteCommandList;
var
 i:integer;
 s:string;
begin
for i:=0 to length(cmdlist)-1 do
if cmdlist[i].cmd<>'' then
begin
 s:=cmdlist[i].cmd;
 if cmdlist[i].p1<>'' then s:=s+' '+cmdlist[i].p1;
 if cmdlist[i].p2<>'' then
 begin
  if cmdlist[i].p1<>'' then s:=s+',' else s:=s+' ';
  s:=s+cmdlist[i].p2;
 end; 
 if cmdlist[i].p3<>'' then s:=s+','+cmdlist[i].p3;
 if cmdlist[i].p4<>'' then s:=s+','+cmdlist[i].p4;
 WriteLine(s+';');
end;
end;

function TPascalShader.CanVarDiscarded(const vn:string;cmdid:integer):boolean;
var
 i:integer;
 si:TSymbolInfo;
 x,y,z,w:boolean;
begin
si:=FindSymbol(vn);
x:=false;y:=false;z:=false;w:=false;
if assigned(si) then
begin
 result:=si.scope<>scoutput;
 for i:=cmdid to length(cmdlist)-1 do
 if cmdlist[i].cmd<>'' then
 begin
  if cmdlist[i].IsVarRead(vn) then
  begin
   result:=false;
   break;
  end;
  if not x and cmdlist[i].IsVarCompWritten(vn,'x') then x:=true;
  if not y and cmdlist[i].IsVarCompWritten(vn,'y') then y:=true;
  if not z and cmdlist[i].IsVarCompWritten(vn,'z') then z:=true;
  if not w and cmdlist[i].IsVarCompWritten(vn,'w') then w:=true;
  if cmdlist[i].IsVarOverWritten(vn) or (x and y and z and w) then
  begin
   result:=true;
   break;
  end;
 end;
end
else
 result:=false;
end;

function TPascalShader.FindNextRead(const vn:string;cmdid:integer):integer;
var
 i:integer;
begin
result:=-1;
for i:=cmdid to length(cmdlist)-1 do
 if (cmdlist[i].cmd<>'') and cmdlist[i].IsVarRead(vn) then
 begin
  result:=i;
  break;
 end; 
end;

function TPascalShader.RemoveDblAssignments:boolean;
var
 i:integer;
begin
result:=false;
for i:=0 to length(cmdlist)-1 do
 if cmdlist[i].cmd<>'' then
 begin
  if CanVarDiscarded(ExtractVarname(cmdlist[i].p1),i+1) then
  begin
   cmdlist[i].cmd:='';
   result:=true;
  end;
 end;
end;

function TPascalShader.RemoveDblCmd:boolean;
var
 i,j:integer;
begin
result:=false;
for i:=0 to length(cmdlist)-1 do
 if (cmdlist[i].cmd<>'') then
 begin
   if cmdlist[i].IsVarWritten(cmdlist[i].p2) or
    cmdlist[i].IsVarWritten(cmdlist[i].p3) or
     cmdlist[i].IsVarWritten(cmdlist[i].p4) then continue;
  if not (length(GetSwizzleMask(cmdlist[i].p1)) in [0,4]) then Continue;
  for j:=i+1 to length(cmdlist)-1 do
  begin
   if (cmdlist[i].cmd=cmdlist[j].cmd) and (cmdlist[i].p2=cmdlist[j].p2) and (cmdlist[i].p3=cmdlist[j].p3) and (cmdlist[i].p4=cmdlist[j].p4) then
   begin
    cmdlist[j].cmd:='MOV';
    cmdlist[j].p2:=cmdlist[i].p1;
    cmdlist[j].p3:='';cmdlist[j].p4:='';
    result:=true;
   end;
   if cmdlist[j].IsVarWritten(cmdlist[i].p1) or cmdlist[j].IsVarWritten(cmdlist[i].p2) or cmdlist[j].IsVarWritten(cmdlist[i].p3) or cmdlist[j].IsVarWritten(cmdlist[i].p4) then break;
  end;
 end; 
end;

function TPascalShader.RemoveAssignments:boolean;
var
 i,j:integer;
 v,v2:string;
begin
result:=false;
for i:=0 to length(cmdlist)-1 do
 if cmdlist[i].cmd='MOV' then
 begin
  if cmdlist[i].p1=cmdlist[i].p2 then
  begin
   cmdlist[i].cmd:='';
   Continue;
  end;
  // Variable mu berschrieben werden
  if not (length(GetSwizzleMask(cmdlist[i].p1)) in [0,4]) then Continue;
  v:=ExtractVarname(cmdlist[i].p1);
  v2:=ExtractVarname(cmdlist[i].p2);
  if FindSymbol(v).scope<>scvar then continue;
  for j:=i+1 to length(cmdlist)-1 do
   if cmdlist[j].cmd<>'' then
  begin
   if ExtractVarname(cmdlist[j].p2)=v then
   begin
    cmdlist[j].p2:=ChangeVarname(cmdlist[j].p2,cmdlist[i].p2);
    result:=true;
   end;
   if ExtractVarname(cmdlist[j].p3)=v then
   begin
    cmdlist[j].p3:=ChangeVarname(cmdlist[j].p3,cmdlist[i].p2);
    result:=true;
   end;
   if ExtractVarname(cmdlist[j].p4)=v then
   begin
    cmdlist[j].p4:=ChangeVarname(cmdlist[j].p4,cmdlist[i].p2);
    result:=true;
   end;
   if cmdlist[j].IsVarWritten(v) or cmdlist[j].IsVarWritten(v2) then break;
  end;
 end;
end;

procedure TPascalShader.FindUsedSymbols;
var
 i:integer;
 si:TSymbolInfo;
begin
for i:=0 to length(cmdlist)-1 do
 if cmdlist[i].cmd<>'' then
 begin
  si:=FindSymbol(ExtractVarname(cmdlist[i].p1));
  if si<>nil then si.used:=true;
  si:=FindSymbol(ExtractVarname(cmdlist[i].p2));
  if si<>nil then si.used:=true;
  si:=FindSymbol(ExtractVarname(cmdlist[i].p3));
  if si<>nil then si.used:=true;
  si:=FindSymbol(ExtractVarname(cmdlist[i].p4));
  if si<>nil then si.used:=true;
 end;
end;


function TPascalShader.OptimizeCMP:boolean;
var
 i,j:integer;
begin
result:=false;
for i:=0 to length(cmdlist)-1 do
 if cmdlist[i].cmd='CMP' then
 begin
  for j:=i+1 to length(cmdlist)-1 do
  begin
  if (cmdlist[j].cmd='CMP') and (cmdlist[j].p1=cmdlist[i].p1) and
   (NegateStr(cmdlist[i].p2)=cmdlist[j].p2) and (cmdlist[j].p4=cmdlist[i].p4) and (cmdlist[j].p1=cmdlist[i].p4) then
  begin
   cmdlist[i].p4:=cmdlist[j].p3;
   cmdlist[j].cmd:='';
   result:=true;
   break;
  end;
  if cmdlist[j].IsVarRead(ExtractVarname(cmdlist[i].p1)) then break;
  if cmdlist[j].IsVarWritten(ExtractVarname(cmdlist[i].p2)) then break;
  if cmdlist[j].IsVarWritten(ExtractVarname(cmdlist[i].p3)) then break;
  if cmdlist[j].IsVarWritten(ExtractVarname(cmdlist[i].p4)) then break;
  end;
 if cmdlist[i].p3=cmdlist[i].p4 then
 begin
  cmdlist[i].cmd:='MOV';
  cmdlist[i].p2:=cmdlist[i].p3;
  cmdlist[i].p3:='';cmdlist[i].p4:='';
 end;
 end;
end;


function TPascalShader.OptimizeSLT:boolean;
var
 i,j,k:integer;
 b:boolean;
begin
result:=false;
for i:=0 to length(cmdlist)-1 do
 if cmdlist[i].cmd='SLT' then
 begin
  j:=FindNextRead(ExtractVarname(cmdlist[i].p1),i+1);
  if j=-1 then continue;
  if (cmdlist[j].cmd<>'SUB') or (cmdlist[j].p3<>cmdlist[i].p1) or (cmdlist[j].p2<>'{0.5,0.5,0.5,0.5}') then continue;
   // Die erste Variable darf nach dem Sub nicht mehr verwendet werden.
  if not CanVarDiscarded(ExtractVarname(cmdlist[i].p1),j+1) then continue;
  // Das Ergebnis des Sub Befehls darf nur fr CMP Befehle als 2. Paramter verwendet werden
  b:=false;
  k:=FindNextRead(ExtractVarname(cmdlist[j].p1),j+1);
  while k<>-1 do
  begin
   if (cmdlist[k].cmd<>'CMP') or ((cmdlist[k].p2<>cmdlist[j].p1) and (cmdlist[k].p2<>negatestr(cmdlist[j].p1)))or
   (cmdlist[k].p3=cmdlist[j].p1) or (cmdlist[k].p4=cmdlist[j].p1)  then
   begin
    b:=true;
    break;
   end;
   if cmdlist[k].IsVarWritten(ExtractVarname(cmdlist[j].p1)) then break;
   k:=FindNextRead(ExtractVarname(cmdlist[j].p1),k+1);
  end;
  if b then continue;
  cmdlist[j].p2:=cmdlist[i].p2;  
  cmdlist[j].p3:=cmdlist[i].p3;
  cmdlist[i].cmd:='';
  result:=true;
 end;
end;


function TPascalShader.OptimizeSUB:boolean;
var
 i:integer;
begin
result:=false;
for i:=0 to length(cmdlist)-1 do
begin
 if (cmdlist[i].cmd='SUB') or (cmdlist[i].cmd='ADD') then
 begin
  if (cmdlist[i].p2='{0,0,0,0}') or (cmdlist[i].p2='-{0,0,0,0}') then
  begin
   cmdlist[i].cmd:='MOV';
   cmdlist[i].p2:=Cmdlist[i].p3;
   cmdlist[i].p3:='';
   result:=true;
  end;
  if (cmdlist[i].p3='{0,0,0,0}') or (cmdlist[i].p3='-{0,0,0,0}') then
  begin
   cmdlist[i].cmd:='MOV';
   cmdlist[i].p3:='';
   result:=true;
  end;
 end;
end;
end;


procedure TPascalShader.OptimizeCommandList;
var
 changed:boolean;
begin
repeat
changed:=false;
changed:=RemoveDblAssignments;
changed:=RemoveDblCmd or changed;
changed:=RemoveAssignments or changed;
changed:=OptimizeCMP or changed;
changed:=OptimizeSLT or changed;
changed:=OptimizeSUB or changed;
until not changed; 
FindUsedSymbols;
end;

procedure TPascalShader.WriteProgram;
begin
codeblock.Write;
OptimizeCommandList;
WriteLine(Header);
WriteVaryingBlock;
WriteParamBlock;
WriteVarBlock;
WriteCommandList;
WriteLine('END');
end;


procedure TPascalShader.InitDefaultVars;
var
 i,j:integer;
begin
InitMatrixList;
AddDefaultSymbol('state.material.ambient',scparam);
AddDefaultSymbol('state.material.diffuse',scparam);
AddDefaultSymbol('state.material.specular',scparam);
AddDefaultSymbol('state.material.emission',scparam);
AddDefaultSymbol('state.material.shininess',scparam);
AddDefaultSymbol('state.material.front.ambient',scparam);
AddDefaultSymbol('state.material.front.diffuse',scparam);
AddDefaultSymbol('state.material.front.specular',scparam);
AddDefaultSymbol('state.material.front.emission',scparam);
AddDefaultSymbol('state.material.front.shininess',scparam);
AddDefaultSymbol('state.material.back.ambient',scparam);
AddDefaultSymbol('state.material.back.diffuse',scparam);
AddDefaultSymbol('state.material.back.specular',scparam);
AddDefaultSymbol('state.material.back.emission',scparam);
AddDefaultSymbol('state.material.back.shininess',scparam);

for i:=0 to 7 do
begin
 AddDefaultSymbol('state.light['+inttostr(i)+'].ambient',scparam);
 AddDefaultSymbol('state.light['+inttostr(i)+'].diffuse',scparam);
 AddDefaultSymbol('state.light['+inttostr(i)+'].specular',scparam);
 AddDefaultSymbol('state.light['+inttostr(i)+'].position',scparam);
 AddDefaultSymbol('state.light['+inttostr(i)+'].attenuation',scparam);
 AddDefaultSymbol('state.light['+inttostr(i)+'].spot.direction',scparam);
 AddDefaultSymbol('state.light['+inttostr(i)+'].half',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].ambient',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].diffuse',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].specular',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].front.ambient',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].front.diffuse',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].front.specular',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].back.ambient',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].back.diffuse',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].back.specular',scparam);
end;
AddDefaultSymbol('state.lightmodel.ambient',scparam);
AddDefaultSymbol('state.lightmodel.scenecolor',scparam);
AddDefaultSymbol('state.lightmodel.front.scenecolor',scparam);
AddDefaultSymbol('state.lightmodel.back.scenecolor',scparam);

for i:=0 to 7 do
begin
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].eye.s',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].eye.t',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].eye.r',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].eye.q',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].object.s',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].object.t',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].object.r',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].object.q',scparam);
end;

AddDefaultSymbol('state.fog.color',scparam);
AddDefaultSymbol('state.fog.params',scparam);

for i:=0 to 5 do
 AddDefaultSymbol('state.clip['+inttostr(i)+'].plane',scparam);

AddDefaultSymbol('state.point.size',scparam);
AddDefaultSymbol('state.point.attenuation',scparam);

for i:=0 to matrixlist.Count-1 do
 for j:=0 to 3 do
  AddDefaultSymbol(matrixlist[i]+'.row['+inttostr(j)+']',scparam);

end;

procedure TPascalShader.InitReservedWords;
begin
with reservedwordlist do
begin
 Add('ABS'); Add('ABS_SAT'); Add('ADD'); Add('ADD_SAT'); Add('ALIAS'); Add('ATTRIB'); Add('CMP'); Add('CMP_SAT'); Add('COS');
 Add('COS_SAT'); Add('DP3'); Add('DP3_SAT'); Add('DP4'); Add('DP4_SAT'); Add('DPH'); Add('DPH_SAT'); Add('DST'); Add('DST_SAT');
 Add('END'); Add('EX2'); Add('EX2_SAT'); Add('FLR'); Add('FLR_SAT'); Add('FRC'); Add('FRC_SAT'); Add('KIL'); Add('LG2');
 Add('LG2_SAT'); Add('LIT'); Add('LIT_SAT'); Add('LRP'); Add('LRP_SAT'); Add('MAD'); Add('MAD_SAT'); Add('MAX'); Add('MAX_SAT');
 Add('MIN'); Add('MIN_SAT'); Add('MOV'); Add('MOV_SAT'); Add('MUL'); Add('MUL_SAT'); Add('OPTION'); Add('OUTPUT'); Add('PARAM');
 Add('POW'); Add('POW_SAT'); Add('RCP'); Add('RCP_SAT'); Add('RSQ'); Add('RSQ_SAT'); Add('SIN'); Add('SIN_SAT'); Add('SCS');
 Add('SCS_SAT'); Add('SGE'); Add('SGE_SAT'); Add('SLT'); Add('SLT_SAT'); Add('SUB'); Add('SUB_SAT'); Add('SWZ'); Add('SWZ_SAT');
 Add('TEMP'); Add('TEX'); Add('TEX_SAT'); Add('TXB'); Add('TXB_SAT'); Add('TXP'); Add('TXP_SAT'); Add('XPD'); Add('XPD_SAT');
 Add('fragment'); Add('program'); Add('result'); Add('state'); Add('texture');
 Add('ADDRESS');  Add('ARL'); Add('EXP');Add('LOG'); Add('vertex');
 Add('cross');add('dot');add('dot3');add('dot4');add('doth');add('floor');
 add('frac');add('exp2');add('log2');add('power');add('normalize');
 add('length');add('sqrlength');add('invlength');add('dist');
 add('sqr');add('sqrt');add('matrixmult3');add('matrixmult4');
 add('vector');add('sat');add('clamp');add('lightatten');add('reflect');
 add('cos');add('sin');add('texture1d');add('texture2d');add('texture3d');
 add('texturecubemap');add('texturerect');add('normalmap');
 add('texture1dproj');add('texture2dproj');add('texture3dproj');
 add('texturecubemapproj');add('texturerectproj');add('lerp');
end;
end;

procedure TPascalShader.InitMatrixList;
var
 i:integer;
begin
matrixlist.Add('state.matrix.modelview');
matrixlist.Add('state.matrix.modelview.inverse');
matrixlist.Add('state.matrix.modelview.transpose');
matrixlist.Add('state.matrix.modelview.invtrans');

for i:=0 to 3 do
begin
 matrixlist.Add('state.matrix.modelview['+inttostr(i)+']');
 matrixlist.Add('state.matrix.modelview['+inttostr(i)+'].inverse');
 matrixlist.Add('state.matrix.modelview['+inttostr(i)+'].transpose');
 matrixlist.Add('state.matrix.modelview['+inttostr(i)+'].invtrans');
end;

matrixlist.Add('state.matrix.projection');
matrixlist.Add('state.matrix.projection.inverse');
matrixlist.Add('state.matrix.projection.transpose');
matrixlist.Add('state.matrix.projection.invtrans');


matrixlist.Add('state.matrix.mvp');
matrixlist.Add('state.matrix.mvp.inverse');
matrixlist.Add('state.matrix.mvp.transpose');
matrixlist.Add('state.matrix.mvp.invtrans');

matrixlist.Add('state.matrix.texture');
matrixlist.Add('state.matrix.texture.inverse');
matrixlist.Add('state.matrix.texture.transpose');
matrixlist.Add('state.matrix.texture.invtrans');

for i:=0 to 7 do
begin
 matrixlist.Add('state.matrix.texture['+inttostr(i)+']');
 matrixlist.Add('state.matrix.texture['+inttostr(i)+'].inverse');
 matrixlist.Add('state.matrix.texture['+inttostr(i)+'].transpose');
 matrixlist.Add('state.matrix.texture['+inttostr(i)+'].invtrans');
end;

matrixlist.Add('state.matrix.program');
matrixlist.Add('state.matrix.program.inverse');
matrixlist.Add('state.matrix.program.transpose');
matrixlist.Add('state.matrix.program.invtrans');

for i:=0 to 31 do
begin
 matrixlist.Add('state.matrix.program['+inttostr(i)+']');
 matrixlist.Add('state.matrix.program['+inttostr(i)+'].inverse');
 matrixlist.Add('state.matrix.program['+inttostr(i)+'].transpose');
 matrixlist.Add('state.matrix.program['+inttostr(i)+'].invtrans');
end;


end;

function TPascalShader.GetSymbolOffset(const v:string):integer;
var
 si:TSymbolInfo;
begin
si:=FindSymbol(v);
if si=nil then result:=-1 else result:=si.offset;
end;


constructor TPascalShader.create(st:TStream);
begin
 inherited create;
 symbollist:=TList.create;
 nodelist:=TList.Create;
 funclist:=TList.create;
 expstack:=TStack.Create;
 matrixlist:=TStringList.create;
 withlist:=TStringList.create;
 reservedwordlist:=TStringList.Create;
 reservedwordlist.CaseSensitive:=false;
 InitReservedWords;
 InitDefaultVars;
 SavedDecimalSeparator:=DecimalSeparator;
 DecimalSeparator:='.';
 self.sc:=TScanner.create(st);
 ParseProgram;
 WriteProgram;
 DecimalSeparator:=SavedDecimalSeparator;
end;

destructor TPascalShader.destroy;
var
 i:integer;
begin
DecimalSeparator:=SavedDecimalSeparator;
for i:=0 to symbollist.count-1 do
 TObject(symbollist[i]).free;
for i:=0 to nodelist.count-1 do
 TObject(nodelist[i]).free;
for i:=0 to funclist.count-1 do
 TObject(funclist[i]).free;
sc.free;
funclist.Free;
nodelist.free;
symbollist.free;
expstack.Free;
matrixlist.free;
withlist.Free;
reservedwordlist.free;
inherited;
end;

// TVertexProgram


function TVertexProgram.WriteExpressionFunc_Cos(exp:TExpression;const target:string=''):string;
const
 c:array[0..4,0..3] of string=(
       ('0.0','0.5','1.0','0.0'),
       ('0.25','-9.0','0.75','0.1591549430'),
       ('24.9808039603','-24.9808039603','-60.1458091736','60.1458091736'),
       ('85.4537887573','-85.4537887573','-64.9393539429','64.9393539429'),
       ('19.7392082214','-19.7392082214','-1.0','1.0'));
var
 p1,r0,r1,r2:string;
begin
exp.CheckParamCount(1);
p1:=exp.paramlist[0].WriteExpression;
FreeTempVar(p1);
p1:=ChangeScalarSwizzle(p1);
r0:=AllocateTempVar;
r1:=AllocateTempVar;
r2:=AllocateTempVar;
WriteCmd('MUL',r1+'.x','{'+c[1,3]+','+c[1,3]+','+c[1,3]+','+c[1,3]+'}',p1);
WriteCmd('FRC',r1+'.y',r1+'.x','');
WriteCmd('SLT',r2+'.x',r1+'.y','{'+c[1,0]+','+c[1,1]+','+c[1,2]+','+c[1,3]+'}');
WriteCmd('SGE',r2+'.yz',r1+'.y','{'+c[1,0]+','+c[1,1]+','+c[1,2]+','+c[1,3]+'}');
WriteCmd('DP3',r2+'.y',r2,'{'+c[4,2]+','+c[4,3]+','+c[4,2]+','+c[4,3]+'}');
WriteCmd('ADD',r0+'.xyz','-'+r1+'.y','{'+c[0,0]+','+c[0,1]+','+c[0,2]+','+c[0,3]+'}');
WriteCmd('MUL',r0,r0,r0);
WriteCmd('MAD',r1,'{'+c[2,0]+','+c[2,1]+','+c[2,0]+','+c[2,1]+'}',r0,'{'+c[1,2]+','+c[2,3]+','+c[2,2]+','+c[2,3]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[3,0]+','+c[3,1]+','+c[3,0]+','+c[3,1]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[3,2]+','+c[3,3]+','+c[3,2]+','+c[3,3]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[4,0]+','+c[4,1]+','+c[4,0]+','+c[4,1]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[4,2]+','+c[4,3]+','+c[4,2]+','+c[4,3]+'}');

if target='' then
begin
 result:=r0;
 WriteCmd('DP3'+exp.sat,r0,r1,'-'+r2);
 FreeTempVar(r1);
 FreeTempVar(r2);
end
else
begin
 WriteCmd('DP3'+exp.sat,target,r1,'-'+r2);
 FreeTempVar(r0);
 FreeTempVar(r1);
 FreeTempVar(r2);
 result:=target;
end;
end;

function TVertexProgram.WriteExpressionFunc_sin(exp:TExpression;const target:string=''):string;
const
 c:array[0..4,0..3] of string=(
       ('0.0','0.5','1.0','0.0'),
       ('0.25','-9.0','0.75','0.1591549430'),
       ('24.9808039603','-24.9808039603','-60.1458091736','60.1458091736'),
       ('85.4537887573','-85.4537887573','-64.9393539429','64.9393539429'),
       ('19.7392082214','-19.7392082214','-1.0','1.0'));
var
 p1,r0,r1,r2:string;
begin
exp.CheckParamCount(1);
p1:=exp.paramlist[0].WriteExpression;
FreeTempVar(p1);
p1:=ChangeScalarSwizzle(p1);
r0:=AllocateTempVar;
r1:=AllocateTempVar;
r2:=AllocateTempVar;
WriteCmd('MAD',r1+'.x','{'+c[1,3]+','+c[1,3]+','+c[1,3]+','+c[1,3]+'}',p1,'{-'+c[1,0]+'}');
WriteCmd('FRC',r1+'.y',r1+'.x','');
WriteCmd('SLT',r2+'.x',r1+'.y','{'+c[1,0]+','+c[1,1]+','+c[1,2]+','+c[1,3]+'}');
WriteCmd('SGE',r2+'.yz',r1+'.y','{'+c[1,0]+','+c[1,1]+','+c[1,2]+','+c[1,3]+'}');
WriteCmd('DP3',r2+'.y',r2,'{'+c[4,2]+','+c[4,3]+','+c[4,2]+','+c[4,3]+'}');
WriteCmd('ADD',r0+'.xyz','-'+r1+'.y','{'+c[0,0]+','+c[0,1]+','+c[0,2]+','+c[0,3]+'}');
WriteCmd('MUL',r0,r0,r0);
WriteCmd('MAD',r1,'{'+c[2,0]+','+c[2,1]+','+c[2,0]+','+c[2,1]+'}',r0,'{'+c[1,2]+','+c[2,3]+','+c[2,2]+','+c[2,3]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[3,0]+','+c[3,1]+','+c[3,0]+','+c[3,1]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[3,2]+','+c[3,3]+','+c[3,2]+','+c[3,3]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[4,0]+','+c[4,1]+','+c[4,0]+','+c[4,1]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[4,2]+','+c[4,3]+','+c[4,2]+','+c[4,3]+'}');

if target='' then
begin
 result:=r0;
 WriteCmd('DP3'+exp.sat,r0,r1,'-'+r2);
 FreeTempVar(r1);
 FreeTempVar(r2);
end
else
begin
 WriteCmd('DP3'+exp.sat,target,r1,'-'+r2);
 FreeTempVar(r0);
 FreeTempVar(r1);
 FreeTempVar(r2);
 result:=target;
end;
end;

function TVertexProgram.WriteExpressionFunc_Lerp(exp:TExpression;const target:string=''):string;
var
 p1,p2,p3:string;
begin
exp.CheckParamCount(3);
p1:=exp.paramlist[0].WriteExpression;
p2:=exp.paramlist[1].WriteExpression;
p3:=exp.paramlist[2].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
FreeTempVar(p3);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('MAD'+exp.sat,result,p3,result,p1);
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('MAD'+exp.sat,target,p3,result,p1);
 result:=target;
end;
end;


function TVertexProgram.WriteExpressionFunction(exp:TExpression;const target:string=''):string;
var
 fname:string;
begin
fname:=lowercase(exp.ident);
if fname='cos' then result:=WriteExpressionFunc_Cos(exp,target) else
if fname='sin' then result:=WriteExpressionFunc_sin(exp,target) else
if fname='lerp' then result:=WriteExpressionFunc_lerp(exp,target)
else result:='';
end;


procedure TVertexProgram.InitDefaultVars;
var
 i:integer;
begin
inherited;
AddDefaultSymbol('vertex.position',scattrib);
AddDefaultSymbol('vertex.weight',scattrib);
AddDefaultSymbol('vertex.weight[0]',scattrib);
AddDefaultSymbol('vertex.weight[1]',scattrib);
AddDefaultSymbol('vertex.weight[2]',scattrib);
AddDefaultSymbol('vertex.weight[3]',scattrib);
AddDefaultSymbol('vertex.normal',scattrib);
AddDefaultSymbol('vertex.color',scattrib);
AddDefaultSymbol('vertex.color.primary',scattrib);
AddDefaultSymbol('vertex.color.secondary',scattrib);
AddDefaultSymbol('vertex.fogcoord',scattrib);
AddDefaultSymbol('vertex.texcoord',scattrib);
for i:=0 to 7 do
 AddDefaultSymbol('vertex.texcoord['+inttostr(i)+']',scattrib);
AddDefaultSymbol('vertex.matrixindex',scattrib);
AddDefaultSymbol('vertex.matrixindex[0]',scattrib);
AddDefaultSymbol('vertex.matrixindex[4]',scattrib);
AddDefaultSymbol('vertex.matrixindex[8]',scattrib);
AddDefaultSymbol('vertex.matrixindex[12]',scattrib);

for i:=0 to 15 do
 AddDefaultSymbol('vertex.attrib['+inttostr(i)+']',scattrib);

AddDefaultSymbol('result.position',scoutput);
AddDefaultSymbol('result.color',scoutput);
AddDefaultSymbol('result.color.primary',scoutput);
AddDefaultSymbol('result.color.secondary',scoutput);
AddDefaultSymbol('result.color.front',scoutput);
AddDefaultSymbol('result.color.front.primary',scoutput);
AddDefaultSymbol('result.color.front.secondary',scoutput);
AddDefaultSymbol('result.color.back',scoutput);
AddDefaultSymbol('result.color.back.primary',scoutput);
AddDefaultSymbol('result.color.back.secondary',scoutput);
AddDefaultSymbol('result.fogcoord',scoutput);
AddDefaultSymbol('result.texcoord',scoutput);

for i:=0 to 7 do
 AddDefaultSymbol('result.texcoord['+inttostr(i)+']',scoutput);

end;

constructor TVertexProgram.create(st:TStream);
begin
AllowDiscard:=false;
SAT_Cmd:=false;
Header:='!!ARBvp1.0';
VaryingType:=scoutput;
inherited;
end;

// TFragmentProgram

function TFragmentProgram.WriteExpressionFunc_Lerp(exp:TExpression;const target:string=''):string;
var
 p1,p2,p3:string;
begin
exp.CheckParamCount(3);
p1:=exp.paramlist[0].WriteExpression;
p2:=exp.paramlist[1].WriteExpression;
p3:=exp.paramlist[2].WriteExpression;
FreeTempVar(p1);
FreeTempVar(p2);
FreeTempVar(p3);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('LRP'+exp.sat,result,p1,p2,p3);
end
else
begin
 WriteCmd('LRP'+exp.sat,target,p1,p2,p3);
 result:=target;
end;
end;

function TFragmentProgram.WriteExpressionFunc_texture(exp:TExpression;const texture:string;const target:string=''):string;
var
 p2,p3:string;
 coord,cmd:string;
begin
if not (length(exp.paramlist) in [2,3]) then
 raise ESyntaxError.Create(tokeninfostr(exp.info)+'Invalid parameter count');
if exp.paramlist[0].info.token<>tkinteger then
 raise ESyntaxError.Create(tokeninfostr(exp.paramlist[0].info)+'Integer expected');

p2:=exp.paramlist[1].WriteExpression;
FreeTempVar(p2);

if length(exp.paramlist)=3 then
begin
 coord:=AllocateTempVar;
 p3:=exp.paramlist[2].WriteExpression;
 FreeTempVar(p3);
 if not SameText(coord,p2) then
  WriteCmd('MOV',coord+'.xyz',p2,'');
 WriteCmd('MOV',coord+'.w',ChangeScalarSwizzle(p3),'');
 cmd:='TXB';
end
else
 begin
 coord:=p2;
 cmd:='TEX';
 end;

FreeTempVar(coord);
if target='' then
 result:=AllocateTempVar
else
 result:=target;
WriteCmd(cmd+exp.sat,result,coord,'texture['+exp.paramlist[0].info.tokenstring+']',texture);

end;

function TFragmentProgram.WriteExpressionFunc_textureproj(exp:TExpression;const texture:string;const target:string=''):string;
var
 p2:string;
begin
exp.CheckParamCount(2);
if exp.paramlist[0].info.token<>tkinteger then
 raise ESyntaxError.Create(tokeninfostr(exp.paramlist[0].info)+'Integer expected');

p2:=exp.paramlist[1].WriteExpression;
FreeTempVar(p2);

if target='' then
 result:=AllocateTempVar
else
 result:=target;
WriteCmd('TXP'+exp.sat,result,p2,'texture['+exp.paramlist[0].info.tokenstring+']',texture);

end;

function TFragmentProgram.WriteExpressionFunc_normalmap(exp:TExpression;const target:string=''):string;
var
 p2,p3:string;
 coord,cmd:string;
begin
if not (length(exp.paramlist) in [2,3]) then
 raise ESyntaxError.Create(tokeninfostr(exp.info)+'Invalid parameter count');
if exp.paramlist[0].info.token<>tkinteger then
 raise ESyntaxError.Create(tokeninfostr(exp.paramlist[0].info)+'Integer expected');

p2:=exp.paramlist[1].WriteExpression;
FreeTempVar(p2);

if length(exp.paramlist)=3 then
begin
 coord:=AllocateTempVar;
 p3:=exp.paramlist[2].WriteExpression;
 FreeTempVar(p3);
 if not SameText(coord,p2) then
  WriteCmd('MOV',coord+'.xyz',p2,'');
 WriteCmd('MOV',coord+'.w',ChangeScalarSwizzle(p3),'');
 cmd:='TXB';
end
else
 begin
 coord:=p2;
 cmd:='TEX';
 end;

FreeTempVar(coord);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd(cmd,result,coord,'texture['+exp.paramlist[0].info.tokenstring+']','2D');
 WriteCmd('MAD'+exp.sat,result,result,'{2.0,2.0,2.0,1.0}','{-1.0,-1.0,-1.0,0.0}');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd(cmd,result,coord,'texture['+exp.paramlist[0].info.tokenstring+']','2D');
 WriteCmd('MAD'+exp.sat,target,result,'{2.0,2.0,2.0,1.0}','{-1.0,-1.0,-1.0,0.0}');
 FreeTempVar(result);
 result:=target;
end;

end;


function TFragmentProgram.WriteExpressionFunction(exp:TExpression;const target:string=''):string;
var
 fname:string;
begin
fname:=lowercase(exp.ident);
if fname='cos' then result:=exp.WriteExpressionFunc_OpCode1s('COS',target) else
if fname='sin' then result:=exp.WriteExpressionFunc_OpCode1s('SIN',target) else
if fname='texture1d' then result:=WriteExpressionFunc_texture(exp,'1D',target) else
if fname='texture2d' then result:=WriteExpressionFunc_texture(exp,'2D',target) else
if fname='texture3d' then result:=WriteExpressionFunc_texture(exp,'3D',target) else
if fname='texturecubemap' then result:=WriteExpressionFunc_texture(exp,'CUBE',target) else
if fname='texturerect' then result:=WriteExpressionFunc_texture(exp,'RECT',target) else
if fname='texture1dproj' then result:=WriteExpressionFunc_textureproj(exp,'1D',target) else
if fname='texture2dproj' then result:=WriteExpressionFunc_textureproj(exp,'2D',target) else
if fname='texture3dproj' then result:=WriteExpressionFunc_textureproj(exp,'3D',target) else
if fname='texturecubemapproj' then result:=WriteExpressionFunc_textureproj(exp,'CUBE',target) else
if fname='texturerectproj' then result:=WriteExpressionFunc_textureproj(exp,'RECT',target) else
if fname='normalmap' then result:=WriteExpressionFunc_normalmap(exp,target) else
if fname='lerp' then result:=WriteExpressionFunc_lerp(exp,target)
else result:='';
end;



procedure TFragmentProgram.InitDefaultVars;
var
 i:integer;
begin
inherited;
AddDefaultSymbol('fragment.color',scattrib);
AddDefaultSymbol('fragment.color.primary',scattrib);
AddDefaultSymbol('fragment.color.secondary',scattrib);
AddDefaultSymbol('fragment.texcoord',scattrib);
for i:=0 to 7 do
 AddDefaultSymbol('fragment.texcoord['+inttostr(i)+']',scattrib);
AddDefaultSymbol('fragment.fogcoord',scattrib);
AddDefaultSymbol('fragment.position',scattrib);

AddDefaultSymbol('result.color',scoutput);
AddDefaultSymbol('result.depth',scoutput);
end;

constructor TFragmentProgram.create(st:TStream);
begin
AllowDiscard:=true;
SAT_Cmd:=true;
CMP_Cmd:=true;
Header:='!!ARBfp1.0';
VaryingType:=scattrib;
inherited;
end;

begin
end.

